# Python-based Risk-Targeted Ground Motion Calculator


#### Author: Andrew Makdisi


Contact: amakdisi@usgs.gov

## Summary

- Presents the development of a Python library for generating Risk-Targeted ground motions, for eventual use as the updated USGS RTGM web tool
- Currently implemented as a Python class called ``riskTargetedGM`` 
- User enters data for a single ground motion hazard curve, uniform hazard return period, conditional collapse probability, fragility beta, and the time horizon over which the target risk is to be calculated 
- Calculation components are highly modular, and involve:
    - Upsampling the ground motion hazard curve
    - Calculating fragility curves given an anchoring IM value and conditional collapse probability
    - Integration of collapse fragility with upsampled hazard curve
    - Iterative risk calculation
- Intermediate calculation information from each iteration is retained
- Class functions for plotting results included


```python
import sys
import os
srcpath = os.path.abspath(os.path.join(os.getcwd(),'..','rtgmpy'))
if srcpath not in sys.path:
    sys.path.append(srcpath)
import matplotlib.pyplot as plt
plt.style.use('makdisi.mplstyle')
```


```python
from RTGMcalculator import*
from HazCurveParser import*
```

#### Getting Hazard Curve Information for some test locations


```python
hazCurvePath = os.path.join(os.getcwd(),'hazard-curves/')
```


```python
# Get IM labels
imLabels = [im for im in os.listdir(hazCurvePath) if '.' not in im]
# Parse site data from a given curves file
im = 'PGA'
curveDF = pd.read_csv(os.path.join(hazCurvePath,im,'curves-truncated.csv'))
siteDF =  curveDF[['name','lon','lat']]
# For a given site, generate a dictinoary for multiple IM hazard curves
site = 'Charleston SC'
```

#### Compiling for a single site


```python
CHS = hazardCurveData(hazCurvePath,siteDF)
```


```python
hazD = CHS.compileSiteHazCurves('Charleston SC',imLabels)
```


```python
haz = {}
# Site metadata
haz['site'] = {}
haz['site']['name'] = site
haz['site']['lat'] = float(siteDF.loc[siteDF['name']==site]['lat'])
haz['site']['lon'] = float(siteDF.loc[siteDF['name']==site]['lon'])
haz['site']['Vs30'] = 760.0

# IM information
haz['hazCurves'] = hazD['IM']
```


```python
hazCurve = hazD['IM']['PGA']
```
import shutil
# plhaDir = 'C:\\Users\\amakdisi\\OneDrive - DOI\\Documents\\CriticalElements\\LTGM\\Data\\pshaFiles'
psha_dir = os.path.join(os.getcwd(),'psha-files')
files = os.listdir(psha_dir)
siteStrings = ['PortlandOR','SanFrancisco','Seattle','Reno','LosAngeles','Boise','LasVegas',
              'Billings','SaltLakeCity','StLouis','Memphis','Paducah','Chicago','Evansville','CharlestonSC',
              'NewYorkCity','Boston']
for sit in siteStrings:
    # Find the Vs30=200m/s dataset
    siteFile = [s for s in files if (sit in s) and ('vs200.0' in s)][0]
#     shutil.copy(os.path.join(psha_dir,siteFile),psha_dir)
    ds0 = xr.open_dataset(os.path.join(psha_dir,siteFile))
## 1. Initializing the ``riskTargetedGM`` class


```python
PrC_tgt = 0.10
beta = 0.60
DEV = riskTargetedGM('PGA',hazCurve,retPeriod_UH=2475,PrC_tgt=PrC_tgt,beta=beta)
```


```python
DEV.compileData()
```





## 2. Hazard Curve Interpolation/Extrpolation ("Upsample")

This is achieved using the ``upsample_hazard_curve`` class method, which takes as its argument:
- The ground motion intensity level at which the maximum and minimum scaling factors are based (typically corresponding initially to a target uniform hazard level (e.g. 975, 1033, 2475-year), which can be calculated using the ``getGMtr`` class method.
- The fragility percentile upon which the upper- and lower-bounds are determinined (e.g. 5% or 10%)
- The standard deviation of the fragility function
- A `linear` flag, which if `True`, performs the upsample using the same methodology as the current, MATLab-based calculator (i.e., linear interpolation/extrapolation in log-log space, and upper and lower bounds based on 10x and 0.1x the IM level.

#### Potential revision to current RTGM web calculator - basing the upper and lower interpolation bounds on target CDF percentiles

Rather than playing around with static, one-size-fits-all scaling factors for the upsampled IM ranges, I decided to simply base them on the fact that we already know what the fragility function is going to look like, i.e. that it will be anchored to a specified value (initially the UH ground motion) at some percentile (5 or 10%). If we want an upsampled range that captures, say, the 0.01th to 99.99th percentiles of the fragility function, then we can simply specify the upper and lower bound values using the following formula:

$\ln(S_{a,p^*}) = \ln(S_{a,p_o}) - \beta\cdot\left[\Phi^{-1}(p_o)-\Phi^{-1}(p^*)\right]$

where $(S_{a,p^*}$ is the spectral amplitude value at our desired percentile $p^*$ (e.g. 0.1% and 99.9% for the lower and upper bounds, respectively), $\ln(S_{a,p_o})$ is the spectral value at which the fragility is anchored (e.g. 1033- or 2475-year values, initially), $p_o$ is the target percentile (e.g. 5% or 10%), $\Phi^{-1}$ is the inverse of the standard normal cumulative distribution function, also known as the *quantile function* or *percent point function*, and $\beta$ is the specified log-standard deviation (e.g. 0.60 or 0.80).

#### Potential revision to current RTGM web calculator - use of spline interpolation/extrapolation in log-space

This seems to result in less conservative exceedance rate estimates at extremely large ground motion levels when compared to the linear extrapolation in log-space tha the current calculator employs. 


```python
riskTargetedGM.upsample_hazard_curve?
```


    [1;31mSignature:[0m
    [0mriskTargetedGM[0m[1;33m.[0m[0mupsample_hazard_curve[0m[1;33m([0m[1;33m
    [0m    [0mself[0m[1;33m,[0m[1;33m
    [0m    [0mtgtGM[0m[1;33m=[0m[1;32mNone[0m[1;33m,[0m[1;33m
    [0m    [0mminCDF[0m[1;33m=[0m[1;36m0.001[0m[1;33m,[0m[1;33m
    [0m    [0mmaxCDF[0m[1;33m=[0m[1;36m0.999[0m[1;33m,[0m[1;33m
    [0m    [0mlinear[0m[1;33m=[0m[1;32mFalse[0m[1;33m,[0m[1;33m
    [0m[1;33m)[0m[1;33m[0m[1;33m[0m[0m
    [1;31mDocstring:[0m
    !
    Class method for interpolating/extrapolating ground motion hazard curve using cubic splines in log-space
    @param tgtGM: Target IM value, used in setting re-sampling bounds (default is UH ground motion level)
    @param minCDF: desired minimum value of corresponding fragility curve
    @param maxCDF: desired maximum value of corresponding fragility curve
    [1;31mFile:[0m      c:\users\amakdisi\projects\rtgm-py-calculator\rtgmpy\rtgmcalculator.py
    [1;31mType:[0m      function
    


### Plotting upsampled hazard curves for all test sites. 

I've included a class method ``plotHazardCurve`` that will generate a plot of the raw data and upsampled hazard curve for the site, and adapted it for multiple sites here using the ``multiHazCurvePlot`` function two cells below.


```python
PrC_tgt = 0.10
beta = 0.60

siteHaz = []
for sit in siteStrings:
    # Find the Vs30=200m/s dataset
    siteFile = [s for s in files if (sit in s) and ('vs200.0' in s)][0]
    ds0 = xr.open_dataset(os.path.join(plhaDir,siteFile))
    TEST = riskTargetedGM(ds0,retPeriod_UH=2475,PrC_tgt=PrC_tgt,beta=beta)
    TEST.upsample_hazard_curve(minCDF=0.001,maxCDF=0.999)
    siteHaz.append(TEST)
col1 = ['black','crimson','royalblue','darkcyan']
col2 = ['gray','orange','mediumorchid','brown']
colList = [col1,col2,col1,col2]
for f in range(4):
    i1 = f*4
    i2 = i1+4
    fig, axes = multiHazCurvePlot(siteHaz[i1:i2],colList=colList[f],addLinear=True)
    for ax in enumerate(axes):
        ax[1].grid(alpha=0.8)
        if ax[0] in [1,3]:
            ax[1].set_ylabel('')

fig,ax = plt.subplots()
fig.set_size_inches(4.75,4.25)
siteHaz[-1].plotHazardCurves(ax)
ax.grid(alpha=0.8)
```


    
![png](SingleCurveRTGM_files/SingleCurveRTGM_21_0.png)
    



    
![png](SingleCurveRTGM_files/SingleCurveRTGM_21_1.png)
    



    
![png](SingleCurveRTGM_files/SingleCurveRTGM_21_2.png)
    



    
![png](SingleCurveRTGM_files/SingleCurveRTGM_21_3.png)
    



    
![png](SingleCurveRTGM_files/SingleCurveRTGM_21_4.png)
    


## 3. Fragility Curves

The fragility curve for a given site is calculated using the ``calc_fragility`` class method. This function generates the collapse CDF from the user-defined target collapse probability (in decimal, e.g. 0.05, 0.10),the corresponding ground motion level (initially the uniform-hazard value, then iterated on), and the beta-value (the lognormal standard deviation, default is 0.80).

Assuming that the specified ground motion level ($S_{a,UH}$) is simply the targeted percentile of the collapse capacity ($p$), the function generates a log-normal distribution, where the median is determined via:

$\ln\mu_{S_a} = \ln(S_{a,UH})-\Phi^{-1}(p)\cdot\beta$

where $\Phi^{-1}$ is the inverse of the standard normal cumulative distribution function, also known as the *quantile function* or *percent point function*, and $\beta$ is the specified standard deviation. 


```python
riskTargetedGM.calc_fragility?
```


    [1;31mSignature:[0m [0mriskTargetedGM[0m[1;33m.[0m[0mcalc_fragility[0m[1;33m([0m[0mself[0m[1;33m,[0m [0manchorIM[0m[1;33m=[0m[1;32mNone[0m[1;33m)[0m[1;33m[0m[1;33m[0m[0m
    [1;31mDocstring:[0m
    !
    Helper method for generating collapse capacity distribution, given a target GM level, anchoring percent collapse, and lognormal standard deviation
    @param anchorIM: Target IM value at which the collapse fragility is anchored (default is the uniform hazard level)
    [1;31mFile:[0m      c:\users\amakdisi\onedrive - doi\documents\criticalelements\buildingcodes\aashto_gm_project\rtgmcalculator.py
    [1;31mType:[0m      function
    



```python
col1 = ['black','crimson','royalblue','darkcyan']
col2 = ['gray','orange','mediumorchid','brown']
colList = [col1,col2,col1,col2]
for f in range(4):
    i1 = f*4
    i2 = i1+4
    fig, axes = multiFragPlot(siteHaz[i1:i2],colList=colList[f]);
    for ax in enumerate(axes):
        ax[1].axhline(1.0,color='gray',ls='--')
        if ax[0] in [1,3]:
            ax[1].set_ylabel('')

fig,ax = plt.subplots()
fig.set_size_inches(4.75,4.25)
siteHaz[-1].plotFragilityCurve(ax,);
ax.axhline(1.0,color='gray',ls='--')
```




    <matplotlib.lines.Line2D at 0x180fc432fa0>




    
![png](SingleCurveRTGM_files/SingleCurveRTGM_24_1.png)
    



    
![png](SingleCurveRTGM_files/SingleCurveRTGM_24_2.png)
    



    
![png](SingleCurveRTGM_files/SingleCurveRTGM_24_3.png)
    



    
![png](SingleCurveRTGM_files/SingleCurveRTGM_24_4.png)
    



    
![png](SingleCurveRTGM_files/SingleCurveRTGM_24_5.png)
    


## 4. Collapse Probability Calculation

1. ``calcRTGM``: class method that calculates the collapse risk iteratively for a target risk and time horizon
    - Handles the initial and subsequent guesses for the ground motion value
    - Tracks and compiles the risk calculations for each iteration
    - Utilizes the ``calculateCollapseRisk`` helper function to calculate the collapse risk for each iteration
    - Generates the class variable ``self.rtgmDict`` containing the IM values, fragility PDFs, the cumulative integral of the hazard curve-fragility PDF product, and the final collapse risk over the specified time horizon, for each iteration.
   
   
   
2. ``calculateCollapseRisk``: helper function that handles the collapse risk calculation for a given "anchoring" ground motion value and time horizon
    - Takes as input the target time horizon and anchoring IM level.
    - Upsamples the hazard curve (``upsample_hazard_curve`` class method) using the anchoring IM level
    - Calculates the collapse fragility (``calc_fragility`` helper function) given the anchoring IM level, and the target collapse probability and standard deviation (already stored as class variables and don't need to be provided as function arguments)
    - Calls the ``integrateCollapseRisk`` helper function to get the final collapse risk values
    - Returns the anchoring IM, fragility PDF, the cumulative integral of the hazard curve-fragility PDF product, and the final collapse risk


3. ``integrateCollapseRisk``: helper function that performs the actual numerical integration of the hazard and fragility curve data
    - Takes as input an array of IM values, the collapse fragility PDF array, and the time horizon
    - Takes the product of the IM hazard curve and collapse PDF, and numerically integrates with respect to the IM values
    - Calculates the final collapse risk over the specified time horizon from the maximum of the hazard curve-fragility curve product integration
    - Returns the im array, the cumulative integral of the hazard curve-fragility curve product, and the final collapse risk


```python
DEV = siteHaz[0]
siteStr = str(np.asarray(DEV.DS['Site'])).split('_')[0]
tgtRisk=0.01
timeYrs=50
DEV.calcRTGM(tgtRisk,timeYrs)
count=0
for it in DEV.rtgmDict['IMguess']:
    count+=1    
    print('Iteration %d: RTGM=%0.5f g'%(count,it))
fig, axs = DEV.plotFullRiskCalc();
fig.suptitle(siteStr,fontsize=16,fontweight='bold')
```

    Iteration 1: RTGM=0.53350 g
    Iteration 2: RTGM=0.39798 g
    Iteration 3: RTGM=0.47264 g
    




    Text(0.5, 0.98, 'PortlandOR')




    
![png](SingleCurveRTGM_files/SingleCurveRTGM_26_2.png)
    


## 5. Example Calc for Fox Plaza

Using the exact same data as obtained from the example calc on [https://earthquake.usgs.gov/designmaps/rtgm/](https://earthquake.usgs.gov/designmaps/rtgm/)


```python
import json
f = open('foxPlaza.json')
FOX = json.load(f)
FOX = FOX['rtgm']
```

#### Initialize the ``riskTargetedGM`` class with the target parameters


```python
from RTGMcalculator import*
os.getcwd()
plhaDir = 'C:\\Users\\amakdisi\\OneDrive - DOI\\Documents\\CriticalElements\\LTGM\\Data\\pshaFiles'
files = os.listdir(plhaDir)
siteStrings = ['PortlandOR','SanFrancisco','Seattle','Reno','LosAngeles','Boise','LasVegas',
              'Billings','SaltLakeCity','StLouis','Memphis','Paducah','Chicago','Evansville','CharlestonSC',
              'NewYorkCity','Boston']
for sit in siteStrings[1:2]:
    # Find the Vs30=200m/s dataset
    siteFile = [s for s in files if (sit in s) and ('vs200.0' in s)][0]
    ds0 = xr.open_dataset(os.path.join(plhaDir,siteFile))
    

# Initialize RTGM Class with target parameters
TR = -50 / np.log(1. - 0.02)
retPeriod_UH=2475
PrC_tgt = 0.10
beta = 0.60
DEV = riskTargetedGM(ds0,retPeriod_UH,PrC_tgt,beta)

# Manually update dataset with Fox Plaza Data
DEV.DS = DEV.DS.drop(['AFEhaz'] )
DEV.DS['PGAhaz']=np.asarray([0.01,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9,1,1.1,1.2,1.4,1.7,2])
DEV.DS['AFEhaz']=(('PGAhaz'),np.asarray([.5696,0.088335,0.02925,0.01229725,0.00564925,0.00275075,
                             0.001385175,0.000733875,0.0003984225,0.0002205625,0.0001235975,
                             0.00006881825,0.0000382493775,0.0000113329875,0.00000134645,
                             0.000000064884]))
DEV.DS['Site'] = 'FoxPlaza'
DEV.getGMtr([DEV.trUH])
```

### a. Upsampling raw hazard curve  data


```python
# Upsample the hazard curve
DEV.upsample_hazard_curve(minCDF=0.01,maxCDF=0.99,tgtGM=DEV.UHGM,linear=True)
fig1, ax1 = plt.subplots()
fig1.set_size_inches(6.5,5.5)
DEV.plotHazardCurves(ax1,plotColor='crimson')
ax1.grid(True,alpha=0.2)
ax1.text(0.05,0.2,'UHGM=%0.3f g'%(float(DEV.UHGM)),
         fontsize=12,fontweight='bold',
         transform=ax1.transAxes,ha='left');
# Plot data from web
imWeb = FOX['upsampledHazardCurve']['xs']
ax1.plot(imWeb,
         FOX['upsampledHazardCurve']['ys'],
        color='gold',ls='--',lw=2,label='Web Tool Data')
ax1.legend(loc='lower left');
```


    
![png](SingleCurveRTGM_files/SingleCurveRTGM_32_0.png)
    


### b. Initial fragility calculation at target UH ground motion level


```python
# Calculate the fragility
im,frPDF,frCDF=DEV.calc_fragility(anchorIM=DEV.UHGM)
fig2, ax2 = plt.subplots()
fig2.set_size_inches(6.5,5.5)
DEV.plotFragilityCurve(ax2,anchorIM=DEV.UHGM,plotColor='blue');
cdfWeb = FOX['iterations'][0]['cdf']
ax2.plot(imWeb,cdfWeb,ls='--',color='orange',lw=2,label='Web Tool Data')
ax2.legend(loc=(0.01,0.3))
```




    <matplotlib.legend.Legend at 0x180fd0fee50>




    
![png](SingleCurveRTGM_files/SingleCurveRTGM_34_1.png)
    


### c. Full Risk Calculation - comparing each component


```python
tgtRisk=0.01
timeYrs=50
DEV.calcRTGM(tgtRisk,timeYrs)
```

#### RTGMs for Fox Plaza (from Web)


```python
count=0
for it in FOX['iterations']:
    count+=1
    print('Iteration %d: RTGM=%0.5f g'%(count,it['rtgm']))
```

    Iteration 1: RTGM=0.79755 g
    Iteration 2: RTGM=0.94316 g
    Iteration 3: RTGM=0.83486 g
    

#### RTGMs for Fox Plaza (from Python)


```python
count=0
for it in DEV.rtgmDict['IMguess']:
    count+=1    
    print('Iteration %d: RTGM=%0.5f g'%(count,it))
```

    Iteration 1: RTGM=0.79756 g
    Iteration 2: RTGM=0.94277 g
    Iteration 3: RTGM=0.83477 g
    

### Comparison of Summary Plots
[https://earthquake.usgs.gov/designmaps/rtgm/](https://earthquake.usgs.gov/designmaps/rtgm/)


```python
fig, axs = DEV.plotFullRiskCalc();
imWeb = FOX['upsampledHazardCurve']['xs']
afeWeb= FOX['upsampledHazardCurve']['ys']

# Convert final CDF data to 50-year probability
for it in FOX['iterations']:
    colCDFweb = it['integral']
    it['integral_T'] = 1-(1-np.asarray(colCDFweb))**DEV.timeYR
    
foxVars = ['hazCurve','cdf','pdf','integrand','integral_T']
colors=['orange','gray','black']
for a in range(1,5):
    # Isolate plot axis object
    ax = axs[a]
    # Plot each iteration
    for it in enumerate(FOX['iterations']):
        plotVar = it[1][foxVars[a]]
        ax.plot(imWeb,plotVar,lw=2,ls='--',color=colors[it[0]],
               label='Web Data - Iter#%d'%(it[0]+1))
    ax.legend(ncol=2)
```


    
![png](SingleCurveRTGM_files/SingleCurveRTGM_42_0.png)
    


## 6. Next Steps
- Testing/finalizing any potential changes to the methodology (i.e. upsampling)
- Adapting for multiple hazard curve inputs (likely will involve a new python class that inherits ``riskTargetedGM``, but accepts hazard curves for multiple IMs and runs the calculation in batch (either looping over each IM hazard curve, or in parallel if computation times necessitate it)
- Discussion on UI development - touch base with Brandon Clayton on Flask (he mentioned it in the most recent code club apparently)
