# 3. Code Usage Guide

Link to example usage jupyter notebooks - `erp-rtgm-calculator/examples`

The core functionality of `rtgmpy` consists of three modules:

- [RiskTargetedGMCalc](../../rtgmpy/RiskTargetedGMCalc.py) contains all of the the computational functions for performing the risk-targeted ground motion calculation for a single hazard curve, with user-defined risk targets.
- [MultiRTGMCalc](../../rtgmpy/MultiRTGMCalc.py) implements the RTGM algorthim for multiple hazard curves within a single site, with user-defined risk targets.
- [BuildingCodeRTGM](../../rtgmpy/BuildingCodeRTGM.py) inherits all the functionality of and performs the same calculations as `MultiRTGMCalc`, but allows the user to instead specify which design guideline document they wish to obtain risk targets for (e.g. `"ASCE7"` or `"AASHTO"`)
 


## 3.1. Input Hazard Curve Data Formats

The user-supplied ground motion hazard curve data must be provided as input to the calculation modules in a python dictionary format. 

For the single ground motion hazard curve input, the python dictionary must be of the following format:

- `'site'`: a nested dictionary with the following site information keys
    - `'name'`: string value denoting the site ID
    - `'lat'`: site latitude in decimal degrees (optional, may be left as `None`)
    - `'lon'`: site longitude in decimal degrees (optional, may be left as `None`)
     - `'Vs30'`: Site Vs30 in m/s (optional, may be left as `None`)
- `'IM'`:   string value denoting the intensity measure (e.g. `'PGA'`, `'SA1P01'`),
- `'iml'`:  1-D numpy array of im values
- `'afe'`:  1-D numpy array of annualized frequencies of exceedance (must be the same length as `'iml'`)

The structure is illustrated in the example dictionary as follows. More example `.json` files can be found in  the `erp-rtgm-calculator/rtgmpy/example-haz-inputs/single-IMs` folder.


```python
hazCurveDict
```




    {'site': {'name': 'Seattle WA', 'lat': 47.6, 'lon': -122.3, 'Vs30': 760.0},
     'IM': 'SA1P0',
     'iml': array([0.0025 , 0.00375, 0.00562, 0.00843, 0.0126 , 0.019  , 0.0284 ,
            0.0427 , 0.064  , 0.096  , 0.144  , 0.216  , 0.324  , 0.486  ,
            0.729  , 1.09   , 1.64   , 2.46   ]),
     'afe': array([3.06510550e-01, 2.70903123e-01, 2.30526701e-01, 1.88283222e-01,
            1.47815364e-01, 1.10590706e-01, 7.98232930e-02, 5.49035455e-02,
            3.62073975e-02, 2.27715926e-02, 1.36468653e-02, 7.76506200e-03,
            4.15349062e-03, 2.04791255e-03, 9.03752778e-04, 3.46282924e-04,
            1.06923716e-04, 2.54856035e-05])}



## 3.2. Calculating risk-targeted ground motion value for a single hazard curve using `RiskTargetedGMCalc`

An instance of the `RiskTargetedGMCalc` is initialized with the following inputs:

- `hazDict`: the dictionary of ground motion hazard curve data, as defined above - default is `None`, which loads an example hazard curve data from one of six CONUS sites in `erp-rtgm-calculator/rtgmpy/example-haz-inputs/short-IMs`
- `retPeriodUH`: Initial uniform hazard return period (e.g. 1033 years for AASHTO, 2475 years for ASCE7)
- `PrC_tgt`: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
- `beta`: Standard deviation of fragility curve (e.g. 0.60)
- `UH_yr`: Time horizon for uniform hazard exceedance rate - used to calculate percent exceedance over time horizon (e.g. 75 yrs for AASHTO, 50 years for ASCE7)

#### ASCE7 Risk Targets

In this example, we use the `SA1P0` hazard curve data from Seattle, and specificy the target collapse risk information as:
- a uniform hazard return period of 2475 years
- conditional collapse risk target of 10%
- lognormal standard deviation of 0.60
- a 50-year time horizon for the uniform hazard


```python
import rtgmpy as rtg

#Initializing the RTGM class for an example Seattle site, for ASCE7 Risk Targets
SEA = rtg.RiskTargetedGMCalc(hazDict=hazCurveDict,
                             retPeriod_UH=2475,
                             PrC_tgt=0.10,
                             beta=0.60,
                            UH_yr=50)
```

#### Class variables
All of the inputs are stored as class variables upon initialization:


```python
initVars=[SEA.siteID,SEA.imLabel,SEA.hazCurve,SEA.trUH,SEA.UH_yr,100*SEA.pctUH,SEA.UHGM,
       100*SEA.PrC_tgt,SEA.beta,]
varLabels=['Site Name: ','Intensity Measure: ','Input Hazard Curve Data: ','Uniform GM Hazard Return Period (yr):',
           'Time Horizon (yr): ','Exceedance Probability (%): ','Interpolated UH ground motion level (g):',
         'Target conditional collapse probability (%)','Lognormal standard deviation of collapse capacity:']
for var in zip(initVars,varLabels):
    print(var[1],'\n\t',var[0])
```

    Site Name:  
    	 Seattle WA
    Intensity Measure:  
    	 SA1P0
    Input Hazard Curve Data:  
    	 {'site': {'name': 'Seattle WA', 'lat': 47.6, 'lon': -122.3, 'Vs30': 760.0}, 'IM': 'SA1P0', 'iml': array([0.0025 , 0.00375, 0.00562, 0.00843, 0.0126 , 0.019  , 0.0284 ,
           0.0427 , 0.064  , 0.096  , 0.144  , 0.216  , 0.324  , 0.486  ,
           0.729  , 1.09   , 1.64   , 2.46   ]), 'afe': array([3.06510550e-01, 2.70903123e-01, 2.30526701e-01, 1.88283222e-01,
           1.47815364e-01, 1.10590706e-01, 7.98232930e-02, 5.49035455e-02,
           3.62073975e-02, 2.27715926e-02, 1.36468653e-02, 7.76506200e-03,
           4.15349062e-03, 2.04791255e-03, 9.03752778e-04, 3.46282924e-04,
           1.06923716e-04, 2.54856035e-05])}
    Uniform GM Hazard Return Period (yr): 
    	 2475
    Time Horizon (yr):  
    	 50
    Exceedance Probability (%):  
    	 1.999932662657944
    Interpolated UH ground motion level (g): 
    	 1.0217256436502353
    Target conditional collapse probability (%) 
    	 10.0
    Lognormal standard deviation of collapse capacity: 
    	 0.6
    

#### Risk Calculation

The iterative calculation is implemented in the `calc_rtgm` method, which requires only the specification:
- `tgtRiskYr`: Target risk of collapse over the specified time horizon (e.g. 0.015 [1.5%] for AASHTO, 0.01 [1%] for ASCE7)
- `timeYrs`: Time horizon over which collapse risk is to be determined (e.g. 75 years for AASHTO, 50 years for ASCE7)
- `errTol`:  (optional) Error below which the calculated RTGM is considered close enough to the target (default is 0.01)
- `maxIter`: (optional) Maximum number of iterations before algorithm terminates (default is 20)

**Note**: Calculation times are from a Windows PC, with an 8-core 2.60GHz processor, with 64GB RAM.


```python
import time as time
start = time.perf_counter()
SEA.calc_rtgm(tgtRiskYr=0.01,timeYrs=50)
end = time.perf_counter()
print('Calculation Time: %0.4f seconds'%(end-start))
```

    Calculation Time: 0.0022 seconds
    

Once the risk calculation is performed, additional class variables are added to the class instance describing the risk targets:


```python
riskVars=[100*SEA.riskTgtYr,SEA.timeYR,SEA.riskTgtAnnual]
riskLabels=['Target collapse probability of over time horizon (%)','Specified time horizon (yr)',
            'Target Annualized collapse risk (yr^-1)']

for var in zip(riskVars,riskLabels):
    print(var[1],'\n\t',var[0])
```

    Target collapse probability of over time horizon (%) 
    	 1.0
    Specified time horizon (yr) 
    	 50
    Target Annualized collapse risk (yr^-1) 
    	 0.000201006717070029
    

#### Summarizing the results using the `data` class variable
The `SEA.data` variable is generated upon class initiation and updated whenever a risk calculation is performed. 
In `.json` format, the final raw data structure is shown below. It's top-level, condensed structure is shown below (**Note:** that the empty brackets are simply condensed lists, containing the numerical array data), and consists of three nested dictionaries:
- `"input"` - This is the same dictionary as the `SEA.hazCurve` class variable, containing all site and hazard curve information provided by the user:
- `"riskTargets"` is a compilation of the fragility and risk target parameters.
- `"RTGM"` provides the detailed results of the risk-targeted calculation - at the top level it includes:
    - `"uhgm"` - the uniform hazard ground motion value (i.e. the initial RTGM guess)
    - `"rtgm"` - the final risk-targeted ground motion
    - `"riskCoeff"` - the ratio between the two
    -`"upsampledHazCurve"`- is a nested dictionary that contains the interpolated/extrapolated hazard curve data used in the calculation algorthim
    - `"iterations"` - is a nested dictionary that contains the full, detailed calculation results for each algorthim iteration. We will explore the structure 


```python
jsview(SEA.data)
```

```javascript
{
   "input": {
      "site": {
         "name": "Seattle WA",
         "lat": 47.6,
         "lon": -122.3,
         "Vs30": 760
      },
      "IM": "SA1P0",
      "iml": [],
      "afe": []
   },
   "riskTargets": {
      "UH_retPeriod": 2475,
      "condCollapseProb": 0.1,
      "beta": 0.6,
      "tgtCollapseRisk_T": 0.01,
      "timeHorizon": 50,
      "riskTgt_Annual": 2.0100671E-4
   },
   "RTGM": {
      "uhgm": 1.0217257,
      "rtgm": 0.9131637,
      "riskCoeff": 0.8937465,
      "upsampledHazCurve": {
         "imlUP": [],
         "afeUP": []
      },
      "iterations": []
   }
}
```

#### Detailed calculation results for each iteration

Within `SEA.data["RTGM"]["iterations"]` we can access the detailed results of each calculation iteration. The structure is shown below for the second iteration (**NOTE:** the empty brackets are simply condensed lists, containing the numerical array data). The key labels correspond to the following variables:
- `"iter"` - The iteration number
- `"rtgmGuess"` - The risk-targeted ground motion value guess used to calculate the collapse risk for this iteration
- `"frCDF"` - The conditional collapse fragility function (i.e. the cumulative distribution function of the collapse capacity)
- `"frPDF"` - The  derivative of the conditional collapse fragility function (i.e. the probability density of the collapse capacity)
- `"hazXpdf"` - The "annual collapse frequency density", or the product of the two terms within the numerical integral (the upsampled hazard curve and the collapse capacity PDF)
- `"colCDF"` - The cumulative annualized collapse frequency 
- `"colCDF_Yr"` - The cumulative collapse probability within the specified time horizon -
- `"colRisk_Annual"` - The calculated annualized collapse risk for this iteration, which is the last value in the `"colCDF"` array
- `"colRisk_Yr"` - The calculated collapse probability over the specified time horizon for this iteration, which is the last value in the `"colCDF_Yr"` array. 

The last two variables are ultimately compared to the risk targets. In this example, we see that the 50-year collapse risk for the second iteration is 1.45%, thus necessitating additional iteration to obtain a design motion targeting the 1.0% in 50 years risk. 

```javascript
{
   "input": {},
   "riskTargets": {},
   "RTGM": {
      "uhgm": 1.0217257,
      "rtgm": 0.9131637,
      "riskCoeff": 0.8937465,
      "upsampledHazCurve": {},
      "iterations": [
         {},
         {
            "iter": 2,
            "rtgmGuess": 0.78169495,
            "frCDF": [],
            "frPDF": [],
            "hazXpdf": [],
            "colCDF": [],
            "colCDF_Yr": [],
            "colRisk_Annual": 2.9117E-4,
            "colRisk_Yr": 0.014455127
         },
         {}
      ]
   }
}
```

The final iteration results can be checked to ensure that the risk target criteria has been met:


```python
finalColRisk_Yr = SEA.data["RTGM"]["iterations"][-1]['colRisk_Yr']
finalColRisk_Annual=SEA.data["RTGM"]["iterations"][-1]['colRisk_Annual']
riskVars=[100*finalColRisk_Yr,finalColRisk_Annual]
riskLabels=['Computed collapse probability of over time horizon (%)','Computed Annualized collapse risk (yr^-1)']

for var in zip(riskVars,riskLabels):
    print(var[1],'\n\t',var[0])
```

    Computed collapse probability of over time horizon (%) 
    	 1.0063461117989392
    Computed Annualized collapse risk (yr^-1) 
    	 0.0002022683419503587
    

### Summary of Component Calculations

The `calc_rtgm` method is largely responsible for compiling/tracking the data from each iteration, and checking the error tolerance, and makes use of several smaller class methods to perform the actual risk calculations themselves. Those methods are:

- `upsample_hazard_curce`: interpolates/extrapolates the user-input IM hazard curve to regularly-log-spaced points
- `calc_fragility`: calculates the collapse fragility for a a given anchoring ground motion value, using the same points as the upsampled hazard curve
- `integrate collapse risk`: numerical integration of the collapse risk given a pair ground motion hazard and fragility curves
- `calc_collapse_risk`: calculates the collapse risk for a single iteration of the anchoring ground motion value

More detailed usage of these component can be found in the jupyter notebooks in  `erp-rtgm-calculator/examples`

## 3.3. Calculating RTGMs for multiple hazard curves at a single site using `MultiRTGMCalc`

This class is very similar to the `RiskTargetedGMCalc` class. However, now our site hazard curve data consists of the dictionary-type structure shown above. The remaining risk target inputs are similar, i.e. uniform hazard return period, conditional collapse probability, lognormal standard deviation.

The additional inputs are the target collapse risk over the time horizon (e.g. 1%), and the specified time horizon (e.g. 50 years)

#### Multi-hazard curve input data

The input dictionary for mutliple hazard curves for a single site is similar in structure to the single-hazard curve format described above, and is illustrated in the example dictionary as follows:


```python
hazCurveDict
```

```javascript
{
   "site": {
      "name": "Seattle WA",
      "lat": 47.6,
      "lon": -122.3,
      "Vs30": 760
   },
   "hazCurves": {
      "PGA": {},
      "SA0P1": {},
      "SA0P5": {},
      "SA1P0": {},
      "SA2P0": {},
      "SA5P0": {},
      "SA10P0": {}
   }
}
```

The individual hazard curve data has the same format as the single-curve dictionary, and can be accessed via:


```python
hazCurveDict['hazCurves']['PGA']
```

```javascript
{
   "iml": [0.00233, 0.0035 , 0.00524, 0.00786, 0.0118 , 0.0177 , 0.0265 ,
        0.0398 , 0.0597 , 0.0896 , 0.134  , 0.202  , 0.302  , 0.454  ,
        0.68   , 1.02   , 1.53   ],
   "afe": [2.91340211e-01, 2.62070273e-01, 2.29738396e-01, 1.94970849e-01,
        1.59294593e-01, 1.24631395e-01, 9.29123075e-02, 6.53963697e-02,
        4.34404430e-02, 2.70791920e-02, 1.57750713e-02, 8.26441978e-03,
        3.86154116e-03, 1.50100497e-03, 4.70732197e-04, 1.10850310e-04,
        1.72898390e-05]
}

```

More example `.json` files can be found in  the `erp-rtgm-calculator/rtgmpy/example-haz-inputs/short-IMs` and `/all-IMs` folders.

#### `MultiRTGMcalc` class initialization and  risk calculation

The multi-hazard curve calculation is performed in a similar manner to the `RiskTargetedGMCalc` class. The class is initiated using the same input parameters, with the only modification being the larger multiple-hazard curve dictionary. The same class variables exist as shown in the previous section.


```python
multSEA = rtg.MultiRTGMcalc(siteDict=hazCurveDict,
                    retPeriod_UH=2475,
                    PrC_tgt=0.10,
                    beta=0.60)
```

The risk calculation is performed using the same input parameters, looping over all IM hazard curves. 

**Note**: Calculation times are from a Windows PC, with an 8-core 2.60GHz processor, with 64GB RAM.


```python
import time
start = time.perf_counter()
multSEA.calc_rtgm_all(tgtRiskYr=0.01,
                     timeYrs=50)
end = time.perf_counter()
print('Calculation Time: %0.4f seconds'%(end-start))
```

    Calculation Time: 0.1243 seconds
    

#### Output Data Structure

The detailed results can be found in the `multSEA.DATA` class variable, which includes the site data, as well as three nested dictionaries:
- `"riskTargets"` contains the same risk target information included in the single-curve `RiskTargetedGMcalc` data dictionary
- `"summary"` contains the `uhgm`, `rtgm`, risk coefficient, and final computed risk values for each intensity measure.
- `"RTGM"` contains the detailed calculation results (similar to the single-curve `RiskTargetedGMcalc` results) for each intensity measure.


```python
multSEA.DATA
```

```javascript
{
   "name": "Seattle WA",
   "lat": 47.6,
   "lon": -122.3,
   "Vs30": 760,
   "riskTargets": {},
   "summary": {},
   "RTGM": {}
}
```


```python
multSEA.DATA['summary']
```

```javascript
{
      "PGA": {
         "uhgm": 0.70976096,
         "rtgm": 0.6600605,
         "riskCoeff": 0.92997575,
         "colRisk_annual": 2.0192591E-4,
         "timePeriod": 50,
         "colRisk_T": 0.010046509
      },
      "SA0P1": {},
      "SA0P5": {},
      "SA1P0": {},
      "SA2P0": {},
      "SA5P0": {},
      "SA10P0": {}
   }
```


```python
multSEA.DATA['RTGM']
```

```javascript
{
      "PGA": {
         "riskTargets": {},
         "rawHaz": {},
         "rtgmCalc": {
            "uhgm": 0.70976096,
            "rtgm": 0.6600605,
            "riskCoeff": 0.92997575,
            "upsampledHazCurve": {
               "imlUP": [],
               "afeUP": []
            },
            "iterations": [
               {
                  "iter": 1,
                  "rtgmGuess": 0.70976096,
                  "frCDF": [],
                  "frPDF": [],
                  "hazXpdf": [],
                  "colCDF": [],
                  "colCDF_Yr": [],
                  "colRisk_Annual": 1.6377619E-4,
                  "colRisk_Yr": 0.008156038
               },
               {},
               {}
            ]
         }
      },
      "SA0P1": {},
      "SA0P5": {},
      "SA1P0": {},
      "SA2P0": {},
      "SA5P0": {},
      "SA10P0": {}
}
```

## 3.4. Calculating RTGMs according to published design guidelines using the `BuildingCodeRTGM` class 

The `BuildingCodeRTGM` class is nearly identical to `MultiRTGMcalc`, the only difference is that the user provides a string input denoting which building code risk targets they would like to use. The current options are `"ASCE7"` and `"AASHTO"`. The output dictionary `bldSEA.DATA` is identical to the `MultiRTGMcalc` class variable, with an additional `"BuildingCode"` key.


```python
bldSEA = rtg.BuildingCodeRTGM(BldgCode='ASCE7',siteDict=hazCurveDict)
bldSEA.calc_rtgm()
```
