# 1. Obtaining Example Hazard Curve Data for Test Sites

For testing/example calculation purposes, data from a series of IM hazard curves, obtained via `nshmp-haz` are available in the `rtgm-py-calculator/development/hazard-curves` directory, for six test sites:
- Charleston, SC,
- Memphis, TN
- Los Angeles, CA
- San Francisco, CA
- Seattle, WA
- Salt Lake City, UT

The hazard curves correspond to 23 IMs, including $PGA$, $PGV$, and $S_a(T)$ ranging from 0.01 to 7.5 seconds.

The raw `nshmp-haz` output is read and processed into a python dictionary using a series of python functions contained in the `rtgm-py-calculator/development/HazCurveParser.py`.

The goal here is to get the hazard curve data in a similar form to how we might process the user input on the actual RTGM calculator web tool. Dictionaries are useful in this regard since they can be read from and written to `.json` format, so that's what the eventual output will look like here.


```python
import sys
import os
srcpath = os.path.abspath(os.path.join(os.getcwd(),'..','rtgmpy'))
devpath = os.path.abspath(os.path.join(os.getcwd(),'..','development'))
for path in [srcpath,devpath]:
    if path not in sys.path:
        sys.path.append(path)
import matplotlib.pyplot as plt
plt.style.use(os.path.join(srcpath,'rtgm.mplstyle'))
from IPython.display import JSON as jsview

from HazCurveParser import*
```

## 1.1. Designating the hazard curve path and getting some basic information about the hazard curve data


```python
# Getting hazard curve data
hazCurvePath = os.path.join(devpath,'hazard-curves/')

# Get IM labels
imLabels = [im for im in os.listdir(hazCurvePath) if '.' not in im]
for im in imLabels:
    print(im)
```

    PGA
    PGV
    SA0P01
    SA0P02
    SA0P03
    SA0P05
    SA0P075
    SA0P1
    SA0P15
    SA0P2
    SA0P25
    SA0P3
    SA0P4
    SA0P5
    SA0P75
    SA10P0
    SA1P0
    SA1P5
    SA2P0
    SA3P0
    SA4P0
    SA5P0
    SA7P5
    

We can also read the `curves.csv` output file for one of the IM runs to get our site information. 


```python
# Parse site data from a given curves file
im = 'PGA'
curveDF = pd.read_csv(os.path.join(hazCurvePath,im,'curves-truncated.csv'))
curveDF
```




<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }

    .dataframe tbody tr th {
        vertical-align: top;
    }

    .dataframe thead th {
        text-align: right;
    }
</style>
<table border="1" class="dataframe">
  <thead>
    <tr style="text-align: right;">
      <th></th>
      <th>name</th>
      <th>lon</th>
      <th>lat</th>
      <th>0.00233</th>
      <th>0.0035</th>
      <th>0.00524</th>
      <th>0.00786</th>
      <th>0.0118</th>
      <th>0.0177</th>
      <th>0.0265</th>
      <th>...</th>
      <th>0.134</th>
      <th>0.202</th>
      <th>0.302</th>
      <th>0.454</th>
      <th>0.68</th>
      <th>1.02</th>
      <th>1.53</th>
      <th>2.3</th>
      <th>3.44</th>
      <th>5.17</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <th>0</th>
      <td>Charleston SC</td>
      <td>-79.93</td>
      <td>32.78</td>
      <td>0.036502</td>
      <td>0.028991</td>
      <td>0.023258</td>
      <td>0.018805</td>
      <td>0.015298</td>
      <td>0.012473</td>
      <td>0.010133</td>
      <td>...</td>
      <td>0.003519</td>
      <td>0.002401</td>
      <td>0.001521</td>
      <td>0.000857</td>
      <td>0.000422</td>
      <td>0.000176</td>
      <td>0.000060</td>
      <td>NaN</td>
      <td>NaN</td>
      <td>NaN</td>
    </tr>
    <tr>
      <th>1</th>
      <td>Memphis</td>
      <td>-90.05</td>
      <td>35.15</td>
      <td>0.100791</td>
      <td>0.081337</td>
      <td>0.063710</td>
      <td>0.048116</td>
      <td>0.035010</td>
      <td>0.024644</td>
      <td>0.016936</td>
      <td>...</td>
      <td>0.003735</td>
      <td>0.002429</td>
      <td>0.001417</td>
      <td>0.000699</td>
      <td>0.000290</td>
      <td>0.000100</td>
      <td>NaN</td>
      <td>NaN</td>
      <td>NaN</td>
      <td>NaN</td>
    </tr>
    <tr>
      <th>2</th>
      <td>Los Angeles CA</td>
      <td>-118.25</td>
      <td>34.05</td>
      <td>1.069517</td>
      <td>0.885232</td>
      <td>0.707611</td>
      <td>0.546561</td>
      <td>0.409559</td>
      <td>0.298737</td>
      <td>0.211865</td>
      <td>...</td>
      <td>0.031147</td>
      <td>0.015563</td>
      <td>0.006911</td>
      <td>0.002507</td>
      <td>0.000703</td>
      <td>0.000139</td>
      <td>0.000017</td>
      <td>NaN</td>
      <td>NaN</td>
      <td>NaN</td>
    </tr>
    <tr>
      <th>3</th>
      <td>San Francisco CA</td>
      <td>-122.40</td>
      <td>37.75</td>
      <td>0.984829</td>
      <td>0.804688</td>
      <td>0.635791</td>
      <td>0.486269</td>
      <td>0.361524</td>
      <td>0.262143</td>
      <td>0.185873</td>
      <td>...</td>
      <td>0.032338</td>
      <td>0.016741</td>
      <td>0.007404</td>
      <td>0.002544</td>
      <td>0.000643</td>
      <td>0.000107</td>
      <td>0.000010</td>
      <td>NaN</td>
      <td>NaN</td>
      <td>NaN</td>
    </tr>
    <tr>
      <th>4</th>
      <td>Seattle WA</td>
      <td>-122.30</td>
      <td>47.60</td>
      <td>0.291340</td>
      <td>0.262070</td>
      <td>0.229738</td>
      <td>0.194971</td>
      <td>0.159295</td>
      <td>0.124631</td>
      <td>0.092912</td>
      <td>...</td>
      <td>0.015775</td>
      <td>0.008264</td>
      <td>0.003862</td>
      <td>0.001501</td>
      <td>0.000471</td>
      <td>0.000111</td>
      <td>0.000017</td>
      <td>NaN</td>
      <td>NaN</td>
      <td>NaN</td>
    </tr>
    <tr>
      <th>5</th>
      <td>Salt Lake City UT</td>
      <td>-111.90</td>
      <td>40.75</td>
      <td>0.178304</td>
      <td>0.141029</td>
      <td>0.108259</td>
      <td>0.080565</td>
      <td>0.058334</td>
      <td>0.041259</td>
      <td>0.028553</td>
      <td>...</td>
      <td>0.005144</td>
      <td>0.003222</td>
      <td>0.001905</td>
      <td>0.000916</td>
      <td>0.000314</td>
      <td>0.000068</td>
      <td>NaN</td>
      <td>NaN</td>
      <td>NaN</td>
      <td>NaN</td>
    </tr>
  </tbody>
</table>
<p>6 rows × 23 columns</p>
</div>




```python
siteDF =  curveDF[['name','lon','lat']]
siteDF
```




<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }

    .dataframe tbody tr th {
        vertical-align: top;
    }

    .dataframe thead th {
        text-align: right;
    }
</style>
<table border="1" class="dataframe">
  <thead>
    <tr style="text-align: right;">
      <th></th>
      <th>name</th>
      <th>lon</th>
      <th>lat</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <th>0</th>
      <td>Charleston SC</td>
      <td>-79.93</td>
      <td>32.78</td>
    </tr>
    <tr>
      <th>1</th>
      <td>Memphis</td>
      <td>-90.05</td>
      <td>35.15</td>
    </tr>
    <tr>
      <th>2</th>
      <td>Los Angeles CA</td>
      <td>-118.25</td>
      <td>34.05</td>
    </tr>
    <tr>
      <th>3</th>
      <td>San Francisco CA</td>
      <td>-122.40</td>
      <td>37.75</td>
    </tr>
    <tr>
      <th>4</th>
      <td>Seattle WA</td>
      <td>-122.30</td>
      <td>47.60</td>
    </tr>
    <tr>
      <th>5</th>
      <td>Salt Lake City UT</td>
      <td>-111.90</td>
      <td>40.75</td>
    </tr>
  </tbody>
</table>
</div>



## 1.2. Initializing the `HazardCurveData` Class 
This requires two input arguments:
- The location of the `nshmp-haz` raw output (i.e. `rtgm-py-calculator/development/hazard-curves`)
- A designation of whether the hazard curve data should be truncated or not (default is `True`)


```python
# For a given site, generate a dictinoary for multiple IM hazard curves
hazCurvePath = os.path.join(devpath,'hazard-curves/')
HAZ = HazardCurveData(hazCurvePath,truncated=True)
```

The `HAZ.get_metadata` method is run as part of the initialization, creating a class variable pandas dataframe of the site information...


```python
HAZ.siteDF
```




<div>
<style scoped>
    .dataframe tbody tr th:only-of-type {
        vertical-align: middle;
    }

    .dataframe tbody tr th {
        vertical-align: top;
    }

    .dataframe thead th {
        text-align: right;
    }
</style>
<table border="1" class="dataframe">
  <thead>
    <tr style="text-align: right;">
      <th></th>
      <th>name</th>
      <th>lon</th>
      <th>lat</th>
    </tr>
  </thead>
  <tbody>
    <tr>
      <th>0</th>
      <td>Charleston SC</td>
      <td>-79.93</td>
      <td>32.78</td>
    </tr>
    <tr>
      <th>1</th>
      <td>Memphis</td>
      <td>-90.05</td>
      <td>35.15</td>
    </tr>
    <tr>
      <th>2</th>
      <td>Los Angeles CA</td>
      <td>-118.25</td>
      <td>34.05</td>
    </tr>
    <tr>
      <th>3</th>
      <td>San Francisco CA</td>
      <td>-122.40</td>
      <td>37.75</td>
    </tr>
    <tr>
      <th>4</th>
      <td>Seattle WA</td>
      <td>-122.30</td>
      <td>47.60</td>
    </tr>
    <tr>
      <th>5</th>
      <td>Salt Lake City UT</td>
      <td>-111.90</td>
      <td>40.75</td>
    </tr>
  </tbody>
</table>
</div>



... and a list of intensity measure labels


```python
HAZ.IMs
```




    ['PGA',
     'PGV',
     'SA0P01',
     'SA0P02',
     'SA0P03',
     'SA0P05',
     'SA0P075',
     'SA0P1',
     'SA0P15',
     'SA0P2',
     'SA0P25',
     'SA0P3',
     'SA0P4',
     'SA0P5',
     'SA0P75',
     'SA10P0',
     'SA1P0',
     'SA1P5',
     'SA2P0',
     'SA3P0',
     'SA4P0',
     'SA5P0',
     'SA7P5']



## 1.3. Obtaining data from a single site and IM

Using the `HAZ.parse_single_hazcurve` method.

Inputs are:
- A string variable indicating the site name (needs to be consistent with one of the names in `HAZ.siteDF`)
- A string variable indicating the desired IM (needs to be consistent with one of the values in `HAZ.IMs`)

Returns a python dictionary containing the site name, intensity measure (`'imt'`), and hazard curve (`'afe'`) values


```python
chPGA = HAZ.parse_single_hazcurve(site='Charleston SC',im='PGA')
```


```python
chPGA
```




    {'name': 'Charleston SC',
     'IM': 'PGA',
     'imt': array([0.00233, 0.0035 , 0.00524, 0.00786, 0.0118 , 0.0177 , 0.0265 ,
            0.0398 , 0.0597 , 0.0896 , 0.134  , 0.202  , 0.302  , 0.454  ,
            0.68   , 1.02   , 1.53   ]),
     'afe': array([3.65022913e-02, 2.89905120e-02, 2.32584616e-02, 1.88053346e-02,
            1.52975039e-02, 1.24729163e-02, 1.01334651e-02, 8.11277180e-03,
            6.35631032e-03, 4.82405608e-03, 3.51851056e-03, 2.40140554e-03,
            1.52115345e-03, 8.56845058e-04, 4.22304777e-04, 1.75579165e-04,
            5.97731843e-05])}



For any dictionary, we can get a summary of the data label information using the `keys()` method


```python
chPGA.keys()
```




    dict_keys(['name', 'IM', 'imt', 'afe'])



And access the values for given key via:


```python
chPGA['imt']
```




    array([0.00233, 0.0035 , 0.00524, 0.00786, 0.0118 , 0.0177 , 0.0265 ,
           0.0398 , 0.0597 , 0.0896 , 0.134  , 0.202  , 0.302  , 0.454  ,
           0.68   , 1.02   , 1.53   ])




```python
chPGA['afe']
```




    array([3.65022913e-02, 2.89905120e-02, 2.32584616e-02, 1.88053346e-02,
           1.52975039e-02, 1.24729163e-02, 1.01334651e-02, 8.11277180e-03,
           6.35631032e-03, 4.82405608e-03, 3.51851056e-03, 2.40140554e-03,
           1.52115345e-03, 8.56845058e-04, 4.22304777e-04, 1.75579165e-04,
           5.97731843e-05])



**Hazard Curve Plot from Dictionary Data**


```python
fig, ax = plt.subplots()
fig.set_size_inches(6.5,4.5)
ax.set_xscale('log')
ax.set_yscale('log')
ax.set_xlabel('Peak Ground Acceleration (g)')
ax.set_ylabel('Annualized Rate of Exceedance ($\mathrm{yr}^{-1}$)')
ax.plot(chPGA['imt'],chPGA['afe'],lw=3,ls='-',marker='o');
```


    
![png](1-ExampleHazardCurveData_files/1-ExampleHazardCurveData_22_0.png)
    


## 1.4. Obtaining multiple (or all available) $IM$ hazard curves for a given site

Using the `Haz.compile_site_hazcurves` method.

Inputs are:
- Site label (similar to previous example)
- List of $IM$ labels (again, all entries should exist in the `self.IMs` list). The default of (`'all'`) parses all available IM data.
- A boolean flag of whether or not the hazard curve data should be `numpy.array`-valued or `list`-valued. `numpy.array` is the default, and advantageous for intermediate calculations, while `list` is more useful for serialization (i.e. saving to a .json file or viewing here)

Returns:
- A slightly more complex dictionary, but still quite usable


```python
CHS = HAZ.compile_site_hazcurves(site='Charleston SC',
                          ims='all',
                          json=True)
```

Using `jsview` should allow you to interact with the json format below a bit here to better understand the dictionary structure.


```python
jsview(CHS)
```




    <IPython.core.display.JSON object>



Let's repeat with `json=False` and plot the hazard curve data:


```python
CHSp = HAZ.compile_site_hazcurves(site='Charleston SC',
                          ims=['PGA','SA0P1','SA0P5',
                               'SA1P0','SA2P0','SA5P0'],
                          json=False)
```

**Hazard Curves Plot from Dictionary Data**


```python
fig, ax = plt.subplots()
fig.set_size_inches(6.5,5.5)
ax.set_xscale('log')
ax.set_yscale('log')
ax.set_xlabel('Ground Motion Intensity (g)')
ax.set_ylabel('Annualized Rate of Exceedance ($\mathrm{yr}^{-1}$)')
# Looping through each im
markers=['o','>','^','s','d','+']
for im in zip(CHSp['IM'],markers):
    imPlot = CHSp['IM'][im[0]]
    ax.plot(imPlot['imt'],imPlot['afe'],lw=2,ls='-',marker=im[1],mfc='w',
           label=im[0]);
    ax.legend(loc='best')
```


    
![png](1-ExampleHazardCurveData_files/1-ExampleHazardCurveData_30_0.png)
    

