# 2. Obtaining Risk-Targeted Ground Motions for a Single $IM$ Hazard Curve

The calculation of risk-targeted ground motion values from a single hazard curve-fragility curve pairing is implemented in the `rtgm-py-calculator/rtgmpy/RiskTargetedGMCalc.py` module, via the `RiskTargetedGMCalc` class. The examples here illustrate:
1. The full, iterative risk-targeted ground motion calculation in its entirety, followed by 
2. Each of the individual component calculations that comprise the iterative RTGM calculation algorithm


```python
import sys
import os
import time
srcpath = os.path.abspath(os.path.join(os.getcwd(),'..','rtgmpy'))
devpath = os.path.abspath(os.path.join(os.getcwd(),'..','development'))
for path in [srcpath,devpath]:
    if path not in sys.path:
        sys.path.append(path)
import matplotlib.pyplot as plt
plt.style.use(os.path.join(srcpath,'rtgm.mplstyle'))
from IPython.display import JSON as jsview

from HazCurveParser import*
from RiskTargetedGMCalc import*
```

## Obtaining example hazard curve input data

For development/example purposes, we obtain the hazard curve data using the methods shown in the previous example notebook, [1-ExampleHazardCurveData.ipynb](./1-ExampleHazardCurveData.ipynb). 


```python
hazCurvePath = os.path.join(devpath,'hazard-curves/')
HAZ = HazardCurveData(hazCurvePath,truncated=True)
LA = HAZ.parse_single_hazcurve(site='Los Angeles CA',im='PGA')
```

The dictionary is intended to be of the same format that we might use to read/process the user-defined input hazard curve data from the web tool itself.


```python
LA
```




    {'name': 'Los Angeles CA',
     'IM': 'PGA',
     'imt': array([0.00233, 0.0035 , 0.00524, 0.00786, 0.0118 , 0.0177 , 0.0265 ,
            0.0398 , 0.0597 , 0.0896 , 0.134  , 0.202  , 0.302  , 0.454  ,
            0.68   , 1.02   , 1.53   ]),
     'afe': array([1.06951702e+00, 8.85232169e-01, 7.07611101e-01, 5.46560856e-01,
            4.09558605e-01, 2.98737382e-01, 2.11865437e-01, 1.44029322e-01,
            9.28045289e-02, 5.58024268e-02, 3.11474655e-02, 1.55630034e-02,
            6.91109093e-03, 2.50735735e-03, 7.03396781e-04, 1.38502525e-04,
            1.72397636e-05])}



## 2.1. Initializing the `RiskTargetedGMCalc` Class


```python
RiskTargetedGMCalc?
```


    Init signature:
    RiskTargetedGMCalc(
        siteName,
        imLabel,
        hazCurve,
        retPeriod_UH,
        PrC_tgt,
        beta,
        UH_yr=50,
    )
    Docstring:     
    !
    Class for calculating risk-targeted ground motion values, anchored to a target conditional collapse probability
    Init docstring:
    !
    @param siteName: String identifying the site for which the hazard curve data is being used
    @param imLabel: String denoting the ground motion intensity measure being considered (e.g. 'PGA', 'SA1P01')
    @param hazCurve: dictionary of ground motion hazard curve data - should consist of two 1-D numpy arrays of the same length with the keys 'imt' and 'afe'
    @param retPeriodUH: Initial target return period (e.g. 1033 for AASHTO, 2475 for ASCE7)
    @param PrC_tgt: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
    @param beta: Standard deviation of fragility curve (e.g. 0.60)
    @param UH_yr: Time horizon for uniform hazard exceedance rate - used to calculate percent exceedance over time horizon (e.g. 75 yrs for AASHTO, 50 years for ASCE7)
    File:           c:\users\amakdisi\projects\rtgm-py-calculator\rtgmpy\risktargetedgmcalc.py
    Type:           type
    Subclasses:        
    


#### Risk-Adjustment for ASCE7

In this example, we use the $PGA$ hazard curve data from Los Angeles, and specificy the target collapse risk information as:
- a uniform hazard return period of 2475 years
- conditional collapse risk target of 10%
- lognormal standard deviation of 0.60
- a 50-year time horizon for the uniform hazard


```python
siteName = 'Los Angeles CA'
imLabel='PGA'
PrC_tgt = 0.10
beta = 0.60
TR_UH = 2475
T_yr = 50

LAX = RiskTargetedGMCalc(siteName=siteName,
                     imLabel = imLabel,
                     hazCurve=LA,
                     retPeriod_UH=TR_UH,
                     PrC_tgt=PrC_tgt,
                     beta=beta,
                    UH_yr=T_yr)
```

#### Initialized class variables

When the `RiskTargetedGM` class is initialized (in this case as the variable `LAX`), the user-defined risk target information is set as class variables (e.g. `LAX.siteID`, `LAX.PrC_tgt`). The uniform hazard ground motion (`LAX.UHGM`) is also interpolated from the hazard curve data and return period, via the `LAX.get_uh_motion` method.


```python
initVars=[LAX.siteID,LAX.imLabel,LAX.hazCurve,LAX.trUH,LAX.UH_yr,100*LAX.pctUH,LAX.UHGM,
       100*LAX.PrC_tgt,LAX.beta,]
varLabels=['Site Name: ',
        'Intensity Measure: ',
        'Input Hazard Curve Data: ',
        'Uniform GM Hazard Return Period (yr):',
        'Time Horizon (yr): ',
         'Exceedance Probability (%): ',
         'Interpolated UH ground motion level (g):',
         'Target conditional collapse probability (%)',
         'Lognormal standard deviation of collapse capacity:'
        ]

for var in zip(initVars,varLabels):
    print(var[1],'\n\t',var[0])
```

    Site Name:  
    	 Los Angeles CA
    Intensity Measure:  
    	 PGA
    Input Hazard Curve Data:  
    	 {'name': 'Los Angeles CA', 'IM': 'PGA', 'imt': array([0.00233, 0.0035 , 0.00524, 0.00786, 0.0118 , 0.0177 , 0.0265 ,
           0.0398 , 0.0597 , 0.0896 , 0.134  , 0.202  , 0.302  , 0.454  ,
           0.68   , 1.02   , 1.53   ]), 'afe': array([1.06951702e+00, 8.85232169e-01, 7.07611101e-01, 5.46560856e-01,
           4.09558605e-01, 2.98737382e-01, 2.11865437e-01, 1.44029322e-01,
           9.28045289e-02, 5.58024268e-02, 3.11474655e-02, 1.55630034e-02,
           6.91109093e-03, 2.50735735e-03, 7.03396781e-04, 1.38502525e-04,
           1.72397636e-05])}
    Uniform GM Hazard Return Period (yr): 
    	 2475
    Time Horizon (yr):  
    	 50
    Exceedance Probability (%):  
    	 1.999932662657944
    Interpolated UH ground motion level (g): 
    	 0.7808818088764742
    Target conditional collapse probability (%) 
    	 10.0
    Lognormal standard deviation of collapse capacity: 
    	 0.6
    

#### The `compile_data` method
This method is called at multiple stages during the calculation, and keeps track of all of the information we might want to eventually provide to the user. It is initially called once the class is initialized and the input data is processed, resulting in the `LAX.data` dictionary (in serializable form). Initially, the data dictionary has two main keys, `'input'` and `'RTGM'`, where `'RTGM'` is appended later on with the full results of the iterative calculation.


```python
jsview(LAX.data)
```




    <IPython.core.display.JSON object>



## 2.2. Full Risk-Targeted Ground Motion Calculation

The iterative calculation is implemented in the `calc_rtgm` method, which requires only the specification of the target probability of collapse of the desired time horizon.


```python
LAX.calc_rtgm?
```


    Signature: LAX.calc_rtgm(tgtRiskYr, timeYrs, errTol=0.01)
    Docstring:
    !
    Class method that iteratively calculates the risk-targeted ground motion value
    @param tgtRiskYr: Target risk of collapse over the specified time horizon (e.g. 0.015 [1.5%] for AASHTO, 0.01 [1%] for ASCE7)
    @param timeYrs: Time horizon over which collapse risk is to be determined (e.g. 75 years for AASHTO, 50 years for ASCE7)
    File:      c:\users\amakdisi\projects\rtgm-py-calculator\rtgmpy\risktargetedgmcalc.py
    Type:      method
    



```python
start = time.perf_counter()
LAX.calc_rtgm(0.01,50)
end = time.perf_counter()
print('Calculation Time: %0.4f seconds'%(end-start))
```

    Calculation Time: 0.0025 seconds
    

#### Additional risk-target class variables


```python
riskVars=[100*LAX.PrC_tgt,LAX.beta,100*LAX.riskTgtYr,LAX.timeYR,LAX.riskTgtAnnual]
riskLabels=['Target conditional collapse probability (%)',
            'Lognormal standard deviation of collapse capacity:',
            'Target collapse probability of over time horizon (%)',
            'Specified time horizon (yr)',
            'Annualized collapse risk (yr^-1)'
        ]

for var in zip(riskVars,riskLabels):
    print(var[1],'\n\t',var[0])
```

    Target conditional collapse probability (%) 
    	 10.0
    Lognormal standard deviation of collapse capacity: 
    	 0.6
    Target collapse probability of over time horizon (%) 
    	 1.0
    Specified time horizon (yr) 
    	 50
    Annualized collapse risk (yr^-1) 
    	 0.000201006717070029
    

#### Final `LAX.data` summary dictionary for a single hazard curve

In `.json` format, the final raw data structure is shown below - I coded this to be pretty similar in structure to what the user gets from the current RTGM web tool.

<img src="ex-figures/jsonView_LAXsingle.png" width=350/>

...which can be saved to a .json file for later viewing/use. For this particule IM, the file size is about 60 KB.


```python
import json
with open('LAX_rtgmPGA.json','w') as jsonFile:
    json.dump(LAX.data,jsonFile)
```

## 2.3. Breakdown of Component Calculations

The `calc_rtgm` method is largely responsible for compiling/tracking the data from each iteration, and checking the error tolerance, and makes use of several smaller class methods to perform the actual risk calculations themselves. Those methods are:

- `upsample_hazard_curce`: interpolates/extrapolates the user-input IM hazard curve to regularly-log-spaced points
- `calc_fragility`: calculates the collapse fragility for a a given anchoring ground motion value, using the same points as the upsampled hazard curve
- `integrate collapse risk`: numerical integration of the collapse risk given a pair ground motion hazard and fragility curves
- `calc_collapse_risk`: calculates the collapse risk for a single iteration of the anchoring ground motion value


```python
LAX = RiskTargetedGMCalc(siteName=siteName,
                     imLabel = imLabel,
                     hazCurve=LA,
                     retPeriod_UH=TR_UH,
                     PrC_tgt=PrC_tgt,
                     beta=beta,
                    UH_yr=T_yr)
```

### 2.3.1 Upsampling Hazard Curve

This function creates a new class variable for the interpolated hazard curve, `self.hazInt`, which is dictionary like, as well as updates the summary dictionary `self.data` with the upsampled hazard curve information.


```python
# Method info/docstring
LAX.upsample_hazard_curve?
```


    Signature:
    LAX.upsample_hazard_curve(
        tgtGM=None,
        minScale=0.1,
        maxScale=10.0,
        linear=True,
    )
    Docstring:
    !
    Class method for log-linear interpolation/extrapolation ground motion hazard curve to higher-resolution arrays
    @param tgtGM: Target IM value, used in setting re-sampling bounds (default is UH ground motion level)
    @param minScale: Lower-bound scaling factor, pegged to uniform hazard value (default is 0.10)
    @param maxScale: Upper-bound scaling factor, pegged to uniform hazard value (default is 10.0)
    File:      c:\users\amakdisi\projects\rtgm-py-calculator\rtgmpy\risktargetedgmcalc.py
    Type:      method
    



```python
# Running the actual method with defaults
LAX.upsample_hazard_curve()
```

#### Updated class variables


```python
LAX.hazInt.keys()
```




    dict_keys(['imtUP', 'afeUP'])




```python
jsview(LAX.data)
```




    <IPython.core.display.JSON object>



(Screenshot of the dictionary in case it's not rendering)

<img src="ex-figures/jsonView_LAXsingleUpSample.png" width=250 height=250 />

#### Plotting the upsampled hazard curve


```python
fig, ax = plt.subplots()
fig.set_size_inches(6.5,5.5)
LAX.plotHazardCurves(ax,plotColor='royalblue')
ax.set_ylim(1e-9)
```




    (1e-09, 10.0)




    
![png](2-RTGM-Single-Hazard-Curve_files/2-RTGM-Single-Hazard-Curve_35_1.png)
    


### 2.3.2 Fragility Curve Calculation

Doesn't update any class variables (hence the term "helper method", versus "class method"), but returns the basic arrays needed to describe the collapse fragility function.


```python
LAX.calc_fragility?
```

    Signature: LAX.calc_fragility(anchorIM=None)
    Docstring:
    !
    Helper method for generating collapse capacity distribution, given a target GM level, anchoring percent collapse, and lognormal standard deviation
    @param anchorIM: Target IM value at which the collapse fragility is anchored (defaults to self.UHGM if None is specified)

    @returns (im, pdf, cdf)
        im: (numpy array) upsampled ground motion amplitudes
        pdf: (numpy array) the probability density function of the collapse capacity
        cdf: (numpy array) the fragility curve (also can be described as the cumulative distribution function of the collapse capacity)            
    File:      c:\users\amakdisi\projects\rtgm-py-calculator\rtgmpy\risktargetedgmcalc.py
    Type:      method
        



```python
im,pdf,cdf=LAX.calc_fragility()
```


```python
fig, ax = plt.subplots()
fig.set_size_inches(6.5,5.5)
LAX.plotFragilityCurve(ax,plotColor='crimson');
```


    
![png](2-RTGM-Single-Hazard-Curve_files/2-RTGM-Single-Hazard-Curve_39_0.png)
    


### 2.3.3. Integration of Hazard and Fragility Curves

Another helper method - here we use the `im` and `pdf` values we just obtained from the fragility calculation, and specify a 50-year time horizon.


```python
LAX.integrate_collapse_risk?
```


    Signature: LAX.integrate_collapse_risk(im, colPDF, numYr)
    Docstring:
    !
    Helper method for calculating the unconditional collapse risk via numerical integration of the ground motion hazard curve and the user-defined collapse capacity
    @param im: array of GM amplitudes, 
    @param colPDF: array of collapse capacity densities, must have the same number of elements as im 
    @param numYr: Time horizon over which collapse risk is to be calculated:
        For NEHRP/ASCE7: 50 years
        For AASHTO: 75 years

    @return (im,hazXpdf,colCDF,colRisk_Y):
        im: (numpy array) upsampled ground motion amplitudes
        hazXpdf: (numpy array) the product of the upsampled hazard curve and collapse capacity PDF
        colCDF: (numpy array) the cumulative collapse risk (over the specified time horizon)
        colRisk_Y: (float) the final collapse risk (over the specified time horizon)
    File:      c:\users\amakdisi\projects\rtgm-py-calculator\rtgmpy\risktargetedgmcalc.py
    Type:      method
    



```python
timeYR = 50
(im,hazXpdf,colCDF,colRisk_Y) = LAX.integrate_collapse_risk(im,pdf,timeYR)
print('Computed %d-yr collapse risk: %0.3f%%'%(timeYR,100*colRisk_Y))
print('Target %d-yr collapse risk: %0.3f%%'%(timeYR,100*0.01))
```

    Computed 50-yr collapse risk: 0.942%
    Target 50-yr collapse risk: 1.000%
    

### 2.3.4. Calculating Collapse Risk for a Single Iteration
Again, another helper function that returns all of the data necessary to compile into that final `LAX.data` summary dictionary. `calculate_collapse_risk` is responsible for assembling all the data needed to run `integrate_collapse_risk`, such as the fragility curve, the upsampled hazard curve, and the collapse risk targets.


```python
LAX.calculate_collapse_risk?
```


    Signature: LAX.calculate_collapse_risk(anchorIM=None, YR=50)
    Docstring:
    !
    Helper method that performs a single risk calculation, based on generating a collapse fragility curve anchored to a user-defined intensity measure, and the class variable self.PrC_tgt
    @param anchorIM: The intensity measure level at which the fragility curve is anchored - default is None, which uses the uniform hazard ground motion level
    @param YR: Used to calculate the risk over a specified time horizont - default is 50 years.

    @return (anchorIM,frCDF,frPDF,hazXpdf,colCDF,colRisk_Y):
        anchorIM: (float) ground motion value used to anchor the fragility curve
        frCDF: (numpy array) the fragility curve (also can be described as the cumulative distribution function of the collapse capacity)
        frPDF: (numpy array) the probability density function of the collapse capacity
        hazXpdf: (numpy array) the product of the upsampled hazard curve and collapse capacity PDF
        colCDF: (numpy array) the cumulative collapse risk (over the specified time horizon)
        colRisk_Y: (float) the final collapse risk (over the specified time horizon)
    File:      c:\users\amakdisi\projects\rtgm-py-calculator\rtgmpy\risktargetedgmcalc.py
    Type:      method
    



```python
(anchorIM,frCDF,frPDF,
 hazXpdf,colCDF,colRisk_Y)=LAX.calculate_collapse_risk()
```

### Summary Plots for All iterations


```python
LAX.calc_rtgm(0.01,50)
fig, ax = LAX.plotFullRiskCalc()
```


    
![png](2-RTGM-Single-Hazard-Curve_files/2-RTGM-Single-Hazard-Curve_47_0.png)
    

