# 3. Obtaining Risk-Targeted Ground Motions for a Set of Hazard Curves


```python
import sys
import os
import time
srcpath = os.path.abspath(os.path.join(os.getcwd(),'..','rtgmpy'))
devpath = os.path.abspath(os.path.join(os.getcwd(),'..','development'))
for path in [srcpath,devpath]:
    if path not in sys.path:
        sys.path.append(path)
import matplotlib.pyplot as plt
plt.style.use(os.path.join(srcpath,'rtgm.mplstyle'))
from IPython.display import JSON as jsview

from HazCurveParser import*
import RiskTargetedGMCalc as rtg
from MultiRTGMcalc import*
```

## Obtaining example hazard curve input data

For development/example purposes, we obtain the hazard curve data using the methods shown in the previous example notebook, [1-ExampleHazardCurveData.ipynb](./1-ExampleHazardCurveData.ipynb). This time, however, we want the ground motion hazard curves for all the available IMs, and instead use the `compile_site_hazcurves` method, rather than `parse_single_hazcurve`.


```python
hazCurvePath = os.path.join(devpath,'hazard-curves/')
HAZ = HazardCurveData(hazCurvePath,truncated=True)
LA = HAZ.compile_site_data(site='Los Angeles CA')
LAjs = HAZ.compile_site_data(site='Los Angeles CA',json=True)
```


```python
jsview(LAjs)
```




    <IPython.core.display.JSON object>



This is the proposed dictionary structure that we may want to use to parse the user-defined hazard curve data.


<img src="ex-figures/jsonView_LAXallIMs.png" width=220/>

## 3.1. Initializing the `MultiRTGMcalc` class

This class is very similar to the `RiskTargetedGMCalc` class that was introduced in [2-RTGM-Single-Hazard-Curve.ipynb](./2-RTGM-Single-Hazard-Curve.ipynb). However, now our site hazard curve data consists of the dictionary-type structure shown above. The remaining risk target inputs are similar, i.e. uniform hazard return period, conditional collapse probability, lognormal standard deviation.

The additional inputs are the target collapse risk over the time horizon (e.g. 1%), and the specified time horizon (e.g. 50 years)


```python
MultiRTGMcalc?
```


    Init signature: MultiRTGMcalc(siteDict, retPeriodUH, PrC_tgt, beta, tgtColRisk_YR, YR)
    Docstring:     
    !
    Class for calculating risk-targeted ground motion values from a set of IM hazard curves, anchored to a target conditional collapse probability
    Init docstring:
    !
    @param siteDict: Dictionary containing all necessary hazard curve data
        keys:   'site': string value denoting the site ID
                'IM':   value is a nested dictionary, the keys of which are the IM labels (e.g. 'PGA', 'SA1P01'), 
                        and the values are dictionaries of ground motion hazard curve data - should consist of two 
                        1-D numpy arrays of the same length with the keys 'imt' and 'afe'
    @param retPeriodUH: Initial target return period (e.g. 1033 for AASHTO, 2475 for ASCE7)
    @param PrC_tgt: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
    @param beta: Standard deviation of fragility curve (e.g. 0.60)
    @param tgtColRisk_Yr: Target collapse risk over the specified time horizon (e.g. 0.015 for AASHTO, 0.01 for ASCE7)
    @param YR: Time horizon for target risk (e.g. 75 years for AASHTO, 50 years for ASCE7)
    File:           c:\users\amakdisi\projects\rtgm-py-calculator\rtgmpy\multirtgmcalc.py
    Type:           type
    Subclasses: 



```python
LAX = MultiRTGMcalc(siteDict=LA,
                    retPeriodUH=2475,
                    PrC_tgt=0.10,
                    beta=0.60,
                    tgtColRisk_YR=0.01,
                    YR=50)
```

## 3.2. Calculate and compile the RTGM data for all IMs
The `calc_rtgm_all` method loops through all IMs and calculates the risk-targeted ground motion by creating a `RiskTargetedGMCalc` instance for each IM. The end result is a list (`LAX.rtgm`) of the summary output dictionary data from each IM.


```python
start = time.perf_counter()
LAX.calc_rtgm_all()
end = time.perf_counter()
print('Calculation Time: %0.4f seconds'%(end-start))
```

    Calculation Time: 0.0252 seconds
    


```python
jsview(LAX.rtgm[0])
```




    <IPython.core.display.JSON object>



The `compile_data` method cleans up this list a bit and turns it into a more structured dictionary that can be written to a json file. 


```python
LAX.compile_data()
```

## 3.3. The overall output `.json` data structure 

Once the data is compiled, the raw output is added to a class variable `LAX.DATA`. This is, once again, a python dictionary that's serializable into .json output file format. There are four main keys to `LAX.DATA`:
- `'site'`
- `'riskTargets'`

<img src="ex-figures/keys_1_2.png" width=220/>

- `'summary'` - the top-line RTGM results for each IM

<img src="ex-figures/keys_3.png" width=300/>


- `'RTGM'` - The detailed RTGM results for each IM (the raw and upsampled hazard curve, the results of each iteration, the intermediate data, etc.)

<img src="ex-figures/keys_4.png" width=300/>


#### Saving to .json

This is quite a bit of data, but the final .json file ends up being around 1.5 MB, which isn't terrible. 


```python
import json
with open('LAX_rtgmALL.json','w') as jsonFile:
    json.dump(LAX.DATA,jsonFile)
```
