# -*- coding: utf-8 -*-
"""
@author: amakdisi@usgs.gov
Child class for calculating risk-targeted ground motions for multiple hazard curve inputs, according to ASCE7 or AASHTO guidelines
Inherited class is MultiRTGMcalc

methods:
    def __init__(self,BldgCode='ASCE7',siteDict=None)
    def compile_data(self)

"""

import csv
import json
import os

import matplotlib.pyplot as plt
import numpy as np

from .MultiRTGMcalc import MultiRTGMcalc
from .RiskTargetedGMCalc import RiskTargetedGMCalc


class BuildingCodeRTGM(MultiRTGMcalc):
    """!
    Class for calculating risk-targeted ground motion values from a set of IM hazard curves, anchored to a target conditional collapse probability
    """

    def __init__(self, BldgCode="ASCE7", siteDict=None):
        """!
        @param BldgCode: String value referencing the construction code document for which to obtain the risk-targeted motions, ('ASCE7' or 'AASHTO')
        @param siteDict: Dictionary containing all necessary hazard curve data - default is `None`, which loads an example hazard curve data from one of six CONUS sites
                        keys:   'site': string value denoting the site ID
                                'IM':   value is a nested dictionary, the keys of which are the IM labels (e.g. 'PGA', 'SA1P01'),
                                        and the values are dictionaries of ground motion hazard curve data - should consist of two
                                        1-D numpy arrays of the same length with the keys 'iml' and 'afe'
        """
        # Defining relevant directories
        self.homeDir = RiskTargetedGMCalc().homeDir
        self.exampleHazDir = RiskTargetedGMCalc().exampleHazDir

        # Assign site data building code tag
        self.siteDict = siteDict
        self.BldgCode = BldgCode

        # Load building code risk targets mapping
        with open(os.path.join(self.homeDir, "BldgCodeTargets.json")) as f:
            self.codeDict = json.load(f)

        # Assign targets as class variables
        try:
            self.riskTargets = self.codeDict[self.BldgCode]
        except (KeyError):
            raise KeyError("BldCode string must be one of %s" % self.codeDict.keys())

        # Initialize inherited class that performs core calculation
        super().__init__(
            self.siteDict,
            retPeriod_UH=self.riskTargets["trUH"],
            PrC_tgt=self.riskTargets["PrC_tgt"],
            beta=self.riskTargets["beta"],
        )

    # Method for defining site data from .csv template (optional), mainly for command-line call
    def read_from_csv(self, csvPath):
        csvData = []
        with open(csvPath) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            for row in csv_reader:
                csvData.append(row)
        # Extract Site Data (if available)
        name = csvData[0][1]
        lat = float(csvData[1][1])
        lon = float(csvData[2][1])
        vs30 = float(csvData[3][1])

        # Adding to new site dictionary
        self.siteDict = {"site": {}}
        for val in zip(["name", "lat", "lon", "Vs30"], [name, lat, lon, vs30]):
            self.siteDict["site"][val[0]] = val[1]

        # Extract hazard curve values
        self.siteDict["hazCurves"] = {}
        numIMs = int(len(csvData[5:]) / 3)

        for im in range(numIMs):
            imRows = csvData[5 + 3 * im : 5 + 3 * im + 3]
            IMT = imRows[0][0]
            # Convert IML and AFE value strings to floats
            IMLs = []
            AFEs = []
            for gm in zip(imRows[1][1:], imRows[2][1:]):
                # Exception handling to avoid reading blank strings
                try:
                    IMLs.append(float(gm[0]))
                    AFEs.append(float(gm[1]))
                except (ValueError):
                    pass
            self.siteDict["hazCurves"][IMT] = {"iml": IMLs, "afe": AFEs}

        # Re-initialize parent class instance
        super().__init__(
            self.siteDict,
            retPeriod_UH=self.riskTargets["trUH"],
            PrC_tgt=self.riskTargets["PrC_tgt"],
            beta=self.riskTargets["beta"],
        )

    # Overloaded data compilation method to add building code tag
    def calc_rtgm(self):
        # Extract Risk Targets
        tgtRiskYr = self.riskTargets["tgtColRisk_T"]
        timeYrs = self.riskTargets["T_yr"]
        super().calc_rtgm_all(tgtRiskYr, timeYrs)
        # super().compile_data()
        newDict = {"BuildingCode": self.BldgCode}
        newDict.update(self.DATA)
        self.DATA = newDict

    def plotRiskCalc(self, imLabel):
        UHyrMap = {"ASCE7": 50, "AASHTO": 75}
        UH_yr = UHyrMap[self.BldgCode]
        if imLabel not in self.DATA["RTGM"]:
            errString = (
                'Intensity Measure "%s" not in hazard curve data.\n\tSelect one of:' % imLabel
                + str(list(self.DATA["RTGM"].keys()))
            )
            raise Exception(errString)
        data = self.DATA["RTGM"][imLabel].copy()
        fig, axs = plt.subplots(5, 1)
        fig.set_size_inches(10, 15)
        # fig.set_tight_layout('tight')
        fig.suptitle(
            self.siteData["name"] + " (%s)" % self.BldgCode,
            fontsize=20,
            fontweight="bold",
            ha="center",
            y=0.91,
        )
        ylabels = [
            "Annual Frequency\nof Exceedance",
            "Conditional Collapse Probability",
            "Conditional Collapse\nProbability Density",
            "Annual Collapse\nFrequency Density",
            "Cumulative %s-Year\nCollapse Probability" % self.riskTargets["T_yr"],
        ]
        varTags = ["hazCurve", "frCDF", "frPDF", "hazXpdf", "colCDF_Yr"]
        titles = [
            "Hazard Curve",
            "Fragility Curves",
            "Derivative of Fragility Curves",
            "Hazard Curve $\\times$\nDerivative of Fragility Curves",
            "Cumulative Integral of Hazard Curve $\\times$\n Derivative of Fragility Curves",
        ]
        labelMap = {
            "PGA": "Peak Ground Acceleration, $PGA$ (g)",
            "PGV": "Peak Ground Velocity, $PGV$ (cm/s)",
        }
        if imLabel not in ["PGA", "PGV"]:
            T_Sa = float(imLabel.split("SA")[-1].replace("P", "."))
            xLabel = "Spectral Acceleration at %0.2f sec, $S_a(%0.2f)$ (g)" % (T_Sa, T_Sa)
        else:
            xLabel = labelMap[imLabel]
        for a in enumerate(axs[0:5]):
            ax = axs[a[0]]
            ax.grid(True, alpha=0.25)
            ax.set_xlabel(xLabel)
            ax.set_ylabel(ylabels[a[0]])
            ax.set_xscale("log")
            if a[0] not in [1, 2, 3, 4]:
                ax.set_yscale("log")
            ax.set_title(titles[a[0]], x=0.01, y=0.05, ha="left", fontsize=14, fontweight="bold")
            # Plot IM hazard curve first
            if a[0] == 0:
                # Draw lines illustrating UHGM
                UHGM = data["rtgmCalc"]["uhgm"]
                trUH = self.riskTargets["trUH"]
                ax.axvline(UHGM, color="gray", ls="--")
                ax.axhline(1 / trUH, color="gray", ls="--")
                # Plot hazard curves
                hazInt = data["rtgmCalc"]["upsampledHazCurve"]
                hazCurve = data["rawHaz"]
                ax.plot(hazInt["imlUP"], hazInt["afeUP"], lw=2.5, color="black", label="Upsampled")
                ax.plot(
                    hazCurve["iml"],
                    hazCurve["afe"],
                    ls="",
                    marker="o",
                    mfc="white",
                    markersize=8,
                    mew=2,
                    color="black",
                    label="Raw Data",
                )
                ax.legend(loc="upper right", fontsize=12)
                # Text for exceedance rate
                (yMin, yMax) = np.log(ax.get_ybound())
                textY = (np.log(1 / self.trUH) - yMin) / (yMax - yMin)
                pctUH = 1 - np.exp(-UH_yr / trUH)
                ax.text(
                    0.01,
                    textY - 0.02,
                    "%d%% Probability of Exceedance in %d Years" % (np.round(100 * pctUH), UH_yr),
                    fontsize=11,
                    transform=ax.transAxes,
                    ha="left",
                    va="top",
                )
            else:
                im = np.asarray(hazInt["imlUP"])
                # Loop through each iteration
                for it in data["rtgmCalc"]["iterations"]:
                    ax.plot(im, it[varTags[a[0]]], lw=2, label="Iteration %d" % it["iter"])
                ax.legend(loc="upper left", fontsize=12)
                if a[0] == 4:
                    ax.axhline(self.riskTargets["tgtColRisk_T"], color="gray", ls="--")
                    # Text for target collapse risk
                    (yMin, yMax) = ax.get_ybound()
                    riskTgtYr = self.riskTargets["tgtColRisk_T"]
                    textY = (riskTgtYr - yMin) / (yMax - yMin)
                    ax.text(
                        0.01,
                        textY - 0.02,
                        "%0.1f%% Probability of Collapse in %d Years"
                        % (100 * riskTgtYr, self.riskTargets["T_yr"]),
                        fontsize=11,
                        transform=ax.transAxes,
                        ha="left",
                        va="top",
                    )
                    ax.legend(loc="lower right", fontsize=12)

        return fig, axs
