# -*- coding: utf-8 -*-
"""
Created on Thu Jan 27 10:32:17 2022

@author: amakdisi

Class for calculating risk-targeted ground motion values from a set of IM hazard curves, anchored to a target conditional collapse probability

methods:
    def __init__(self,siteDict=None,retPeriodUH=2475,PrC_tgt=0.10,beta=0.60,tgtColRisk_YR=0.01,YR=50)
    def calc_rtgm_all(self)
    def summarize_im_data(self,rtDict)
    def compile_data(self)
"""

import copy
import json
import os
import random

import numpy as np
import pandas as pd

from .RiskTargetedGMCalc import RiskTargetedGMCalc


class MultiRTGMcalc(object):
    """!
    Class for calculating risk-targeted ground motion values from a set of IM hazard curves, anchored to a target conditional collapse probability
    """

    def __init__(self, siteDict=None, retPeriod_UH=2475, PrC_tgt=0.10, beta=0.60, UH_yr=50):
        """!
        @param siteDict: Dictionary containing all necessary hazard curve data - default is `None`, which loads an example hazard curve data from one of six CONUS sites
            keys:   'site': string value denoting the site ID
                    'IM':   value is a nested dictionary, the keys of which are the IM labels (e.g. 'PGA', 'SA1P01'),
                            and the values are dictionaries of ground motion hazard curve data - should consist of two
                            1-D numpy arrays of the same length with the keys 'iml' and 'afe'
        @param retPeriod_UH: Initial target return period (e.g. 1033 for AASHTO, 2475 for ASCE7)
        @param PrC_tgt: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
        @param beta: Standard deviation of fragility curve (e.g. 0.60)
        @param UH_yr: Time horizon for uniform hazard exceedance rate - used to calculate percent exceedance over time horizon (e.g. 75 yrs for AASHTO, 50 years for ASCE7)

        """
        # Defining relevant directories
        self.homeDir = RiskTargetedGMCalc().homeDir
        self.exampleHazDir = RiskTargetedGMCalc().exampleHazDir

        # Set site info and hazard curve class variables
        self.siteDict = siteDict
        if self.siteDict:
            self.siteData = self.siteDict["site"]
            self.hazData = self.siteDict["hazCurves"]
            self.IMs = list(self.hazData)
        else:
            # Get the example .json hazard curve file names
            exFiles = [file for file in os.listdir(self.exampleHazDir) if ".json" in file]
            # Pick one at random and load into a dictionary
            hazEx = random.choice(exFiles)
            with open(os.path.join(self.exampleHazDir, hazEx)) as f:
                data = json.load(f)
            # Assign dictionary components to class variables
            # self.siteData = data["site"]
            self.siteData = data["site"]
            self.hazData = data["hazCurves"]
            self.IMs = list(self.hazData)
            # Populate new siteDict class variable with combined data
            self.siteDict = {}
            self.siteDict["site"] = self.siteData
            self.siteDict["hazCurves"] = self.hazData

        # Risk target parameters
        self.trUH = retPeriod_UH
        self.UH_yr = UH_yr
        self.pctUH = 1 - np.exp(-self.UH_yr / self.trUH)
        self.PrC_tgt = PrC_tgt
        self.beta = beta

    def calc_rtgm_all(self, tgtRiskYr, timeYrs, errTol=0.01, maxIter=20):
        """!
        Class method that performs iterative RTGM calculation for all IMs
        @param tgtRiskYr: Target risk of collapse over the specified time horizon (e.g. 0.015 [1.5%] for AASHTO, 0.01 [1%] for ASCE7)
        @param timeYrs: Time horizon over which collapse risk is to be determined (e.g. 75 years for AASHTO, 50 years for ASCE7)
        @param errTol:  (optional) Error below which the calculated RTGM is considered close enough to the target (default is 0.01)
        @param maxIter: (optional) Maximum number of iterations before algorithm terminates (default is 20)
        """

        self.errTol = errTol
        self.maxIter = maxIter

        # Setting the risk target and time horizon as class variables
        self.riskTgtYr = tgtRiskYr
        self.timeYR = timeYrs

        # Calculate targete annualized risk
        self.riskTgtAnnual = -np.log(1 - self.riskTgtYr) / timeYrs

        self.rtgm = []
        for im in self.IMs:
            hazDict_im = {
                "site": self.siteData,
                "IM": im,
                "iml": np.asarray(self.hazData[im]["iml"]),
                "afe": np.asarray(self.hazData[im]["afe"]),
            }
            rtIM = RiskTargetedGMCalc(
                hazDict=hazDict_im,
                retPeriod_UH=self.trUH,
                PrC_tgt=self.PrC_tgt,
                beta=self.beta,
                UH_yr=self.timeYR,
            )
            rtIM.calc_rtgm(self.riskTgtYr, timeYrs=self.timeYR)
            self.rtgm.append(rtIM.data)
        self.errTol = rtIM.errTol

        # Compile data
        self.compile_data()

    def summarize_im_data(self, rtDict):
        """!
        Helper method that extracts summary data from a given RTGM calc and returns a dictionary
        @param rtDict: the full self.data dictionary for a given IM calculation (from the riskTargetedGM class)
        """
        # Pull the summary RT values for each IM
        keys = {"uhgm", "rtgm", "riskCoeff"}
        summ = {k: rtDict["RTGM"][k] for k in rtDict["RTGM"] if k in keys}
        # Add annualized, T-year risk, and time horizon
        summ["colRisk_annual"] = rtDict["RTGM"]["iterations"][-1]["colRisk_Annual"]
        summ["timePeriod"] = self.timeYR
        summ["colRisk_T"] = rtDict["RTGM"]["iterations"][-1]["colRisk_Yr"]
        return summ

    def compile_data(self):
        """!
        Class method that compiles data from all RTGM calculations into a dictionary - can be serialized into json format
        """
        # Throwing site information into the DATA dictionary
        self.DATA = {"site": self.siteData}

        # Summary of risk targets
        self.DATA["riskTargets"] = {}
        targets = zip(
            ["UH_retPeriod", "condCollapseProb", "beta", "tgtCollapseRisk_T", "timeHorizon"],
            [self.trUH, self.PrC_tgt, self.beta, self.riskTgtYr, self.timeYR],
        )
        for tgt in targets:
            self.DATA["riskTargets"][tgt[0]] = tgt[1]

        # Add in annualized
        self.DATA["riskTargets"]["riskTgt_Annual"] = -np.log(1 - self.riskTgtYr) / self.timeYR

        # RTGM Summary
        self.DATA["summary"] = {}
        self.DATA["RTGM"] = {}

        # Making sure we actually have the rtgm data first
        try:
            dummy = self.rtgm
        except (AttributeError):
            self.calc_rtgm_all()

        for im in self.rtgm:
            self.DATA["summary"][im["input"]["IM"]] = self.summarize_im_data(im)
            # Extract detailed calculation results for each IM
            imdict = copy.deepcopy(im)
            # imdict["input"].pop("siteID")
            imdict["input"].pop("IM")
            imdict["rawHaz"] = imdict["input"]
            imdict["rawHaz"].pop("site")
            imdict.pop("input")
            imdict["rtgmCalc"] = imdict["RTGM"]
            imdict.pop("RTGM")

            # Append to RTGM key
            self.DATA["RTGM"][im["input"]["IM"]] = imdict

        # Also create a pandas dataframe summary
        self.summaryDF = pd.DataFrame(self.DATA["summary"]).transpose()
