# -*- coding: utf-8 -*-
"""
@author: amakdisi@usgs.gov
Class for calculating risk-targeted ground motion values, anchored to a target conditional collapse probability

methods:
    def __init__(self,siteName,imLabel,hazCurve,retPeriod_UH,PrC_tgt,beta,UH_yr=50)
    def compile_data(self)
    def get_uh_motion(self, trGM)
    def upsample_hazard_curve(self,tgtGM=None,minScale=0.10,maxScale=10.0,linear=True)
    def calc_fragility(self,anchorIM=None)
    def integrate_collapse_risk(self,im,colPDF,numYr)
    def calculate_collapse_risk(self,anchorIM=None,YR=None)
    def calc_rtgm(self,tgtRiskYr,timeYrs,errTol=0.01)
"""

import json
import os
import random
import sys

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy
import scipy.interpolate as interpolate
import scipy.stats as st

from cycler import cycler
from matplotlib.ticker import ScalarFormatter


mpl.rcParams["axes.prop_cycle"] = cycler(
    "color", ["royalblue", "crimson", "darkcyan", "orange", "mediumorchid", "gray"]
)


class RiskTargetedGMCalc(object):
    """!
    Class for calculating risk-targeted ground motion value, anchored to a target conditional collapse probability
    """

    # Defining relevant directories
    homeDir = os.path.dirname(os.path.realpath(__file__))
    exampleHazDir = os.path.join(homeDir, "example-haz-inputs", "short-IMs")

    def __init__(self, hazDict=None, retPeriod_UH=2475, PrC_tgt=0.10, beta=0.60, UH_yr=50):
        """!
        @param hazDict: dictionary of ground motion hazard curve data - default is `None`, which loads an example hazard curve data from one of six CONUS sites
            keys:   'site': a nested dictionary with the following site information keys
                            'name': string value denoting the site ID
                            'lat': site latitude in decimal degrees (optional, may be left as None)
                            'lon': site longitude in decimal degrees (optional, may be left as None)
                            'Vs30': Site Vs30 in m/s (optional, may be left as None)
                    'IM':   string value denoting the intensity measure (e.g. 'PGA', 'SA1P01'),
                    'iml':  1-D numpy array of im values
                    'afe':  1-D numpy array of annualized frequencies of exceedance (must be the same length as 'iml')
        @param retPeriodUH: Initial target return period (e.g. 1033 for AASHTO, 2475 for ASCE7)
        @param PrC_tgt: Target conditional collapse fragility percentile (in decimal form, e.g. 0.05 for AASHTO, 0.10 for ASCE7)
        @param beta: Standard deviation of fragility curve (e.g. 0.60)
        @param UH_yr: Time horizon for uniform hazard exceedance rate - used to calculate percent exceedance over time horizon (e.g. 75 yrs for AASHTO, 50 years for ASCE7)
        """

        # Processing site input
        if type(hazDict) == dict:
            self.hazCurve = hazDict
            for key in ["iml", "afe"]:
                self.hazCurve[key] = np.array(hazDict[key])
        else:
            # Get the example .json hazard curve file names
            exFiles = [
                file for file in os.listdir(RiskTargetedGMCalc.exampleHazDir) if ".json" in file
            ]
            # Pick one at random and load into a dictionary
            hazEx = random.choice(exFiles)
            with open(os.path.join(RiskTargetedGMCalc.exampleHazDir, hazEx)) as f:
                data = json.load(f)
            # Extract random iml out of it
            self.hazCurve = {"site": data["site"]}
            imLabel = random.choice(list(data["hazCurves"].keys()))
            self.hazCurve["IM"] = imLabel
            # Assign hazard curve arrays to class dictionary
            for key in ["iml", "afe"]:
                self.hazCurve[key] = np.asarray(data["hazCurves"][imLabel][key])

        # Assign site id and imLabel
        self.siteID = self.hazCurve["site"]["name"]
        self.imLabel = self.hazCurve["IM"]

        # Data summary is initially an empty dict
        self.data = {}

        # Uniform hazard values
        self.trUH = retPeriod_UH
        self.UH_yr = UH_yr
        self.pctUH = 1 - np.exp(-self.UH_yr / self.trUH)

        # Collapse risk targets
        self.PrC_tgt = PrC_tgt
        self.beta = beta

        # #
        self.get_uh_motion(trGM=self.trUH)

        # # Compile data
        self.compile_data()

    def compile_data(self):
        # Compile input data
        self.data = {
            "input": {
                "site": self.hazCurve["site"],
                "IM": self.imLabel,
                "iml": list(self.hazCurve["iml"]),
                "afe": list(self.hazCurve["afe"]),
            }
        }
        # Initialize empty dictionary for RTGM results
        self.data["RTGM"] = {}
        if self.UHGM:
            self.data["RTGM"]["uhgm"] = self.UHGM

        return

    def get_uh_motion(self, trGM):
        """!
        Class method for estimating ground motion value from a hazard curve for a given return period
        @param trGM: Target return period
        """
        # if not trsGM:
        #     trsGM = [self.trUH]
        # if 'trsGM' in self.DS.variables:
        #     self.DS = self.DS.drop(['trsGM','gmTR'])

        # Update Uniform hazard values
        self.trUH = trGM
        self.pctUH = 1 - np.exp(-self.UH_yr / self.trUH)

        f = interpolate.interp1d(np.log(self.hazCurve["afe"]), np.log(self.hazCurve["iml"]))
        self.UHGM = np.exp(f(np.log(1 / np.asarray(trGM))))

        # If we've initialized the data summary, add the uhgm value to it
        if (self.data) and ("RTGM" in self.data):
            self.data["RTGM"]["uhgm"] = self.UHGM

    def upsample_hazard_curve(self, tgtGM=None, minScale=0.10, maxScale=10.0, linear=True):
        """!
        Class method for log-linear interpolation/extrapolation ground motion hazard curve to higher-resolution arrays
        @param tgtGM: Target IM value, used in setting re-sampling bounds (default is UH ground motion level)
        @param minScale: Lower-bound scaling factor, pegged to uniform hazard value (default is 0.10)
        @param maxScale: Upper-bound scaling factor, pegged to uniform hazard value (default is 10.0)
        """

        # # Minimum scaling factor is based on 0.01st percentile
        # minScale = np.exp(self.beta*(st.norm.ppf(minCDF)-st.norm.ppf(self.PrC_tgt)))

        # # Maximum scaling factor is based on 99.99th percentile
        # maxScale = np.exp(self.beta*(st.norm.ppf(maxCDF)-st.norm.ppf(self.PrC_tgt)))

        # Min and max bounds based on user-defined scaling factors
        if not tgtGM:
            tgtGM = self.UHGM

        minGM = tgtGM * minScale
        maxGM = tgtGM * maxScale

        minGM = min(minGM, self.hazCurve["iml"].min())
        maxGM = max(maxGM, self.hazCurve["iml"].max())

        gmInt = np.exp(np.arange(np.log(minGM), np.log(maxGM), np.log(1.05)))
        if gmInt.max() <= maxGM:
            gmInt = np.append(gmInt, maxGM)

        # if not linear:
        #     # Interpolation using splines
        #     tck = interpolate.interp1d(np.log(self.DS['PGAhaz']),np.log(self.DS['AFEhaz']),kind='cubic',
        #                                     fill_value='extrapolate')
        # else:
        #     # Interpolation using linear extrapolate (to re-create current RTGM)
        #     tck = interpolate.interp1d(np.log(self.DS['PGAhaz']),np.log(self.DS['AFEhaz']),kind='linear',
        #                                     fill_value='extrapolate')

        # Interpolation using linear extrapolate (to re-create current RTGM)
        tck = interpolate.interp1d(
            np.log(self.hazCurve["iml"]),
            np.log(self.hazCurve["afe"]),
            kind="linear",
            fill_value="extrapolate",
        )
        AFEint = np.exp(tck(np.log(gmInt)))

        # Updating the upsampled data (if necessary)
        self.hazInt = {"imlUP": gmInt, "afeUP": AFEint}
        # Adding to the master data dictionary (as list data)
        if (self.data) and ("RTGM" in self.data):
            self.data["RTGM"]["upsampledHazCurve"] = {}
            for k in self.hazInt:
                self.data["RTGM"]["upsampledHazCurve"][k] = self.hazInt[k].tolist()

    def calc_fragility(self, anchorIM=None):
        """!
        Helper method for generating collapse capacity distribution, given a target GM level, anchoring percent collapse, and lognormal standard deviation
        @param anchorIM: Target IM value at which the collapse fragility is anchored (defaults to self.UHGM if None is specified)

        @returns (im, pdf, cdf)
            im: (numpy array) upsampled ground motion amplitudes
            pdf: (numpy array) the probability density function of the collapse capacity
            cdf: (numpy array) the fragility curve (also can be described as the cumulative distribution function of the collapse capacity)
        """
        if not anchorIM:
            anchorIM = self.UHGM
        # Extract im array (perform upsampling if needed)
        try:
            im = self.hazInt["imlUP"]
        except (AttributeError):
            self.upsample_hazard_curve()
            im = self.hazInt["imlUP"]
        # Calculate log-median from percentile value
        logMed = np.log(anchorIM) - st.norm.ppf(self.PrC_tgt) * self.beta
        # Calculate pdf and cdf
        pdf = st.norm.pdf(np.log(im), logMed, self.beta)
        cdf = st.norm.cdf(np.log(im), logMed, self.beta)
        # pdf = np.gradient(cdf,im)
        return (im, pdf, cdf)

    def integrate_collapse_risk(self, im, colPDF, numYr):
        """!
        Helper method for calculating the unconditional collapse risk via numerical integration of the ground motion hazard curve and the user-defined collapse capacity
        @param im: array of GM amplitudes,
        @param colPDF: array of collapse capacity densities, must have the same number of elements as im
        @param numYr: Time horizon over which collapse risk is to be calculated:
            For NEHRP/ASCE7: 50 years
            For AASHTO: 75 years

        @return (im,hazXpdf,colCDF,colRisk_Y):
            im: (numpy array) upsampled ground motion amplitudes
            hazXpdf: (numpy array) the product of the upsampled hazard curve and collapse capacity PDF
            colCDF: (numpy array) the cumulative collapse risk (over the specified time horizon)
            colRisk_Y: (float) the final collapse risk (over the specified time horizon)
        """
        # Calculating annual collapse frequency density (product of hazard curve and collapse PDF)
        hazXpdf = colPDF * self.hazInt["afeUP"]
        # Numerical integration
        colCDF = scipy.integrate.cumtrapz(hazXpdf, np.log(im), initial=0)
        colProb = scipy.integrate.trapz(hazXpdf, np.log(im))
        # Calculate collapse risk over specified time horizon
        colRisk_Y = 1 - (1 - colProb) ** numYr
        return (im, hazXpdf, colCDF, colRisk_Y)

    def calculate_collapse_risk(self, anchorIM=None, YR=50):
        """!
        Helper method that performs a single risk calculation, based on generating a collapse fragility curve anchored to a user-defined intensity measure, and the class variable self.PrC_tgt
        @param anchorIM: The intensity measure level at which the fragility curve is anchored - default is None, which uses the uniform hazard ground motion level
        @param YR: Used to calculate the risk over a specified time horizont - default is 50 years.

        @return (anchorIM,frCDF,frPDF,hazXpdf,colCDF,colRisk_Y):
            anchorIM: (float) ground motion value used to anchor the fragility curve
            frCDF: (numpy array) the fragility curve (also can be described as the cumulative distribution function of the collapse capacity)
            frPDF: (numpy array) the probability density function of the collapse capacity
            hazXpdf: (numpy array) the product of the upsampled hazard curve and collapse capacity PDF
            colCDF: (numpy array) the cumulative collapse risk (over the specified time horizon)
            colRisk_Y: (float) the final collapse risk (over the specified time horizon)
        """
        if not anchorIM:
            anchorIM = self.UHGM

        # Calculating collapse fragility
        (im, frPDF, frCDF) = self.calc_fragility(anchorIM)
        # Calculating collapse risk
        (im, hazXpdf, colCDF, colRisk_Y) = self.integrate_collapse_risk(
            im=self.hazInt["imlUP"], colPDF=frPDF, numYr=YR
        )
        return (anchorIM, frCDF, frPDF, hazXpdf, colCDF, colRisk_Y)

    def calc_rtgm(self, tgtRiskYr, timeYrs, errTol=0.01, maxIter=20):
        """!
        Class method that iteratively calculates the risk-targeted ground motion value
        @param tgtRiskYr: Target risk of collapse over the specified time horizon (e.g. 0.015 [1.5%] for AASHTO, 0.01 [1%] for ASCE7)
        @param timeYrs: Time horizon over which collapse risk is to be determined (e.g. 75 years for AASHTO, 50 years for ASCE7)
        @param errTol:  (optional) Error below which the calculated RTGM is considered close enough to the target (default is 0.01)
        @param maxIter: (optional) Maximum number of iterations before algorithm terminates (default is 20)
        """
        self.errTol = errTol
        self.maxIter = maxIter

        # Setting the risk target and time horizon as class variables
        self.riskTgtYr = tgtRiskYr
        self.timeYR = timeYrs

        # Calculate targete annualized risk
        self.riskTgtAnnual = -np.log(1 - self.riskTgtYr) / timeYrs

        # Update data class variable with risk targets
        self.data["riskTargets"] = {}
        targets = zip(
            ["UH_retPeriod", "condCollapseProb", "beta", "tgtCollapseRisk_T", "timeHorizon"],
            [self.trUH, self.PrC_tgt, self.beta, self.riskTgtYr, self.timeYR],
        )
        for tgt in targets:
            self.data["riskTargets"][tgt[0]] = tgt[1]

        # Add in annualized
        self.data["riskTargets"]["riskTgt_Annual"] = -np.log(1 - self.riskTgtYr) / self.timeYR

        # Create empty list to track results over all iterations
        self.rtgmData = []
        colRisk_Y = None
        colCDF = None
        for nIter in range(maxIter):
            # First iteration - UHS value is set as first guess
            if nIter == 0:
                self.get_uh_motion(self.trUH)
                IMval = self.UHGM
            # Second iteration, IM guess is based on ratio between target and 1st-guess collapse risks
            elif nIter == 1:
                # Checking against target
                err = colCDF[-1] / self.riskTgtAnnual
                if np.abs(err - 1) <= errTol:
                    break
                # Next guess
                IMval = err * IMval
            # 3+ iteration - Log-linearly interpolate afterwards
            else:
                # Checking against target
                err = colCDF[-1] / self.riskTgtAnnual
                if np.abs(err - 1) <= errTol:
                    break
                # Otherwise, we interpolate
                # extracting annualized collapse risk and rtgm guesses
                colRisks = np.asarray([it["colRisk_Annual"] for it in self.rtgmData])
                IMguesses = np.asarray([it["rtgmGuess"] for it in self.rtgmData])
                xp = np.log(colRisks)
                yp = np.log(IMguesses)
                f = interpolate.interp1d(xp, yp)
                IMval = np.exp(f(np.log(self.riskTgtAnnual)))

            # Calculate the collapse risk
            (IMval, frCDF, frPDF, hazXpdf, colCDF, colRisk_Y) = self.calculate_collapse_risk(
                anchorIM=IMval, YR=self.timeYR
            )
            colCDF_Yr = 1 - (1 - colCDF) ** self.timeYR
            calcRiskAnnual = colCDF[-1]

            # Create results dictionary for each iteration
            dictLabels = [
                "iter",
                "rtgmGuess",
                "frCDF",
                "frPDF",
                "hazXpdf",
                "colCDF",
                "colCDF_Yr",
                "colRisk_Annual",
                "colRisk_Yr",
            ]
            dictVars = [
                nIter + 1,
                IMval,
                frCDF,
                frPDF,
                hazXpdf,
                colCDF,
                colCDF_Yr,
                calcRiskAnnual,
                colRisk_Y,
            ]
            iterDict = {}
            for data in zip(dictLabels, dictVars):
                if type(data[1]) is np.ndarray:
                    iterDict[data[0]] = list(data[1])
                else:
                    iterDict[data[0]] = data[1]
            self.rtgmData.append(iterDict)

        # Update the summary data dictionary
        self.data["RTGM"]["rtgm"] = self.rtgmData[-1]["rtgmGuess"]
        self.data["RTGM"]["riskCoeff"] = self.data["RTGM"]["rtgm"] / self.data["RTGM"]["uhgm"]

        # Do some quick re-ordering
        hazInt = self.data["RTGM"]["upsampledHazCurve"]
        self.data["RTGM"].pop("upsampledHazCurve")
        self.data["RTGM"]["upsampledHazCurve"] = hazInt

        self.data["RTGM"]["iterations"] = self.rtgmData
        # Add the final RTGM value to the summary
        # Reorder
        self.data = {
            "input": self.data["input"],
            "riskTargets": self.data["riskTargets"],
            "RTGM": self.data["RTGM"],
        }

    def plotHazardCurves(self, ax, plotColor="black"):
        # Draw lines illustrating UHGM
        ax.axvline(self.UHGM, color="gray", ls="--")
        ax.axhline(1 / self.trUH, color="gray", ls="--")

        # Plot interpolated curves
        ax.plot(
            self.hazInt["imlUP"], self.hazInt["afeUP"], lw=3, color=plotColor, label="Upsampled"
        )
        ax.set_xscale("log")
        ax.set_yscale("log")
        ax.set_xlabel("PGA (g)")
        ax.set_ylabel(r"Annualized Exceedance ($\mathrm{yr}^{-1}$)")

        # Plot raw hazard curve data
        ax.plot(
            self.hazCurve["iml"],
            self.hazCurve["afe"],
            ls="",
            marker="o",
            mfc="white",
            color=plotColor,
            label="Raw Hazard Curve Data",
        )

        # Setting axis limits
        afeMin = float(self.hazCurve["afe"].min())
        limLB = np.floor(np.log10(afeMin)) - 2

        # Setting axis limits
        afeMax = float(self.hazCurve["afe"].max())
        limUB = np.floor(np.log10(afeMax)) + 1

        ax.set_ybound(10 ** (limLB), 10**limUB)
        # Set Axis title
        ax.text(
            0.98,
            0.98,
            str(np.asarray(self.data["input"]["site"]["name"])).split("_")[0],
            fontweight="bold",
            fontsize=14,
            transform=ax.transAxes,
            ha="right",
            va="top",
        )
        ax.legend(loc="lower left")

    def plotFragilityCurve(self, ax, anchorIM=None, plotColor="black"):
        if not anchorIM:
            anchorIM = self.UHGM
        # Calculate fragility curve
        (im, pdf, cdf) = self.calc_fragility(anchorIM)
        ax.grid(alpha=0.3)
        ax.set_xscale("log")
        ax.axvline(anchorIM, color="black", ls="--")
        ax.axhline(self.PrC_tgt, color="black", ls="--")

        ax.plot(im, cdf, lw=3, color=plotColor)
        # ax.set_xlim(1e-1,50)
        ax.set_ylabel("Probability of Collapse")
        ax.set_xlabel("Spectral Acceleration, $S_a$ (g)")
        ax.set_ylim(-0.05, 1.05)
        # Label for target collapse risk
        yTa = 1.0 / 1.1 * (self.PrC_tgt + 0.05)
        ax.text(
            0.02,
            yTa + 0.02,
            "%0.1f%% Target\nConditional Collapse\nProbability" % (100 * self.PrC_tgt),
            fontsize=12,
            transform=ax.transAxes,
        )
        # Label for UHGM
        scale = np.diff(np.log10(ax.get_xlim()))
        xmin = np.log10(ax.get_xlim())[0]
        xTa = (np.log10(anchorIM) - xmin) / scale
        ax.text(
            xTa + 0.01,
            0.015,
            "Uniform Hazard \nGM (%d-year)" % (self.trUH),
            fontsize=12,
            transform=ax.transAxes,
            ha="left",
        )
        # Set Axis title
        ax.text(
            0.02,
            0.98,
            self.siteID,
            fontweight="bold",
            fontsize=14,
            transform=ax.transAxes,
            ha="left",
            va="top",
        )
        return ax

    def plotFullRiskCalc(self):
        data = self.rtgmData.copy()
        fig, axs = plt.subplots(5, 1)
        fig.set_size_inches(10, 15)
        fig.set_tight_layout("tight")
        ylabels = [
            "Annual Frequency\nof Exceedance",
            "Conditional Collapse Probability",
            "Conditional Collapse\nProbability Density",
            "Annual Collapse\nFrequency Density",
            "Cumulative %s-Year\nCollapse Probability" % self.timeYR,
        ]

        varTags = ["hazCurve", "frCDF", "frPDF", "hazXpdf", "colCDF_Yr"]
        titles = [
            "Hazard Curve",
            "Fragility Curves",
            "Derivative of Fragility Curves",
            "Hazard Curve $\\times$\nDerivative of Fragility Curves",
            "Cumulative Integral of Hazard Curve $\\times$\n Derivative of Fragility Curves",
        ]
        for a in enumerate(axs[0:5]):
            ax = axs[a[0]]
            ax.grid(True, alpha=0.25)
            ax.set_xlabel("Spectral Acceleration (g)")
            ax.set_ylabel(ylabels[a[0]])
            ax.set_xscale("log")
            if a[0] not in [1, 2, 3, 4]:
                ax.set_yscale("log")
            ax.set_title(titles[a[0]], x=0.01, y=0.05, ha="left", fontsize=14, fontweight="bold")
            # Plot IM hazard curve first
            if a[0] == 0:
                # Draw lines illustrating UHGM
                ax.axvline(self.UHGM, color="gray", ls="--")
                ax.axhline(1 / self.trUH, color="gray", ls="--")
                # Plot hazard curves
                ax.plot(
                    self.hazInt["imlUP"],
                    self.hazInt["afeUP"],
                    lw=2.5,
                    color="black",
                    label="Upsampled",
                )
                ax.plot(
                    self.hazCurve["iml"],
                    self.hazCurve["afe"],
                    ls="",
                    marker="o",
                    mfc="white",
                    markersize=8,
                    mew=2,
                    color="black",
                    label="Raw Data",
                )
                ax.legend(loc="upper right", fontsize=12)
                # Text for exceedance rate
                (yMin, yMax) = np.log(ax.get_ybound())
                textY = (np.log(1 / self.trUH) - yMin) / (yMax - yMin)
                ax.text(
                    0.01,
                    textY - 0.02,
                    "%d%% Probability of Exceedance in %d Years"
                    % (np.round(100 * self.pctUH), self.UH_yr),
                    fontsize=11,
                    transform=ax.transAxes,
                    ha="left",
                    va="top",
                )
            else:
                im = np.asarray(self.hazInt["imlUP"])
                # Loop through each iteration
                for it in self.rtgmData:
                    ax.plot(im, it[varTags[a[0]]], lw=2, label="Iteration %d" % it["iter"])
                ax.legend(loc="upper left", fontsize=12)
                if a[0] == 4:
                    ax.axhline(0.01, color="gray", ls="--")
                    # Text for target collapse risk
                    (yMin, yMax) = ax.get_ybound()
                    textY = (self.riskTgtYr - yMin) / (yMax - yMin)
                    ax.text(
                        0.01,
                        textY - 0.02,
                        "%d%% Probability of Collapse in %d Years"
                        % (np.round(100 * self.riskTgtYr), self.timeYR),
                        fontsize=11,
                        transform=ax.transAxes,
                        ha="left",
                        va="top",
                    )
                    ax.legend(loc="lower right", fontsize=12)
        return fig, axs


# Simple function to plot
def multiHazCurvePlot(siteList, colList=None, addLinear=False):
    fig, axes = plt.subplots(2, 2)
    fig.set_size_inches(10.5, 8.5)
    if not colList:
        cols = ["black", "crimson", "royalblue", "darkcyan"]
    else:
        cols = colList
    axRet = []
    for ax in enumerate(axes.flatten()):
        siteList[ax[0]].plotHazardCurves(ax[1], plotColor=cols[ax[0]])
        if addLinear:
            rtgm = siteList[ax[0]]
            rtgm.upsample_hazard_curve(linear=True)
            ax[1].plot(
                rtgm.DS["gmInt"], rtgm.DS["AFEint"], color="gray", ls="--", lw=2, label="Linear"
            )
            ax[1].legend()
        axRet.append(ax[1])
    return fig, axRet


def multiFragPlot(siteList, colList=None):
    fig, axes = plt.subplots(2, 2)
    fig.set_size_inches(10.5, 8.5)
    if not colList:
        cols = ["black", "crimson", "royalblue", "darkcyan"]
    else:
        cols = colList
    axRet = []
    for ax in enumerate(axes.flatten()):
        siteList[ax[0]].plotFragilityCurve(ax[1], plotColor=cols[ax[0]])
        axRet.append(ax[1])
    return fig, axRet
